import altair as alt
from millify import millify, prettify


def no_axis():
    """ Returns an invisible Altair axis. """
    return alt.Axis(domain=False, grid=False, ticks=False, labels=False, title=" ")


def format_number(value):
    """ Format number to be easily read by humans. """
    if value > 1000000:
        return millify(value, precision=2)
    return prettify(value)


def transform_ratio(value):
    """ Transformation that takes ratios and applies a function that preserves equal distances to origin (1) 
    for similar relationships, eg. a ratio of 2 (twice the size) is at the same distance of 1 (same size) as 
    0.5 (half the size).
    Read: 'how many times larger or smaller than reference disparity'."""

    if value >= 1:
        return value - 1
    else:
        return 1 - 1 / value


def calculate_chart_size_from_elements(header_size, element_size, number_elements):
    """ Calculates the total size for a chart based on the desired size for the header (title, axis, etc.), and the 
    size and number of elements (lines, columns). """

    return header_size + (element_size * number_elements)


def to_list(user_input):
    return user_input if (isinstance(user_input, list)) else [user_input]



import math
import altair as alt
import pandas as pd

from aequitas.plot.bubble_disparity_chart import get_disparity_bubble_chart_components
from aequitas.plot.bubble_metric_chart import get_metric_bubble_chart_components

from aequitas.plot.commons.legend import draw_legend

from aequitas.plot.commons.style.classes import Title, Metric_Axis, Chart_Title
from aequitas.plot.commons.style.text import FONT
from aequitas.plot.commons.style.sizes import Concat_Chart
from aequitas.plot.commons import initializers as Initializer

# Altair 2.4.1 requires that all chart receive a dataframe, for charts that don't need it
# (like most annotations), we pass the following dummy dataframe to reduce the complexity of the resulting vega spec.
DUMMY_DF = pd.DataFrame({"a": [1, 1], "b": [0, 0]})


def __get_chart_sizes(chart_width):
    """Calculates the widths of the disparity and metric charts that make-up the concatenated chart.
    The individual widths are calculated based on the provided desired overall chart width."""

    chart_sizes = dict(
        disparity_chart_width=0.5 * chart_width, metric_chart_width=0.5 * chart_width
    )
    return chart_sizes


def draw_chart_title(chart_title, chart_width):
    """Draws chart titles."""

    title = (
        alt.Chart(DUMMY_DF)
        .mark_text(
            align="center",
            fill=Title.font_color,
            fontSize=Title.font_size,
            fontWeight=Title.font_weight,
            font=FONT,
            tooltip="",
        )
        .encode(
            x=alt.value(chart_width / 2),
            y=alt.value(Title.margin_top),
            text=alt.value(chart_title),
        )
    )

    return title


def plot_concatenated_bubble_charts(
    disparity_df,
    metrics_list,
    attribute,
    fairness_threshold=1.25,
    chart_height=None,
    chart_width=Concat_Chart.full_width,
    accessibility_mode=False,
):
    """Draws a concatenation of the disparity bubble chart and the metric values bubble chart,
    of the selected metrics for a given attribute.

    :param disparity_df: a dataframe generated by the Aequitas Bias class
    :type disparity_df: pandas.core.frame.DataFrame
    :param metrics_list: a list of the metrics of interest
    :type metrics_list: list
    :param attribute: an attribute to plot
    :type attribute: str
    :param fairness_threshold: a value for the maximum allowed disparity, defaults to 1.25
    :type fairness_threshold: float, optional
    :param chart_height: a value (in pixels) for the height of the chart
    :type chart_height: int, optional
    :param chart_width: a value (in pixels) for the width of the chart
    :type chart_width: int, optional
    :param accessibility_mode: a switch for the display of more accessible visual elements, defaults to False
    :type accessibility_mode: bool, optional

    :return: the full disparities chart
    :rtype: Altair chart object
    """

    (
        plot_table,
        metrics,
        ref_group,
        global_scales,
        chart_height,
        chart_width,
        selection,
    ) = Initializer.prepare_bubble_chart(
        disparity_df,
        metrics_list,
        attribute,
        fairness_threshold,
        chart_height,
        chart_width,
        Concat_Chart,
        accessibility_mode,
    )

    chart_sizes = __get_chart_sizes(chart_width)

    # TITLES
    disparity_title = draw_chart_title(
        "Disparities", chart_sizes["disparity_chart_width"]
    )
    metric_title = draw_chart_title(
        "Absolute Values", chart_sizes["metric_chart_width"]
    )

    # DISPARITY CHART
    disparity_chart = (
        (
            get_disparity_bubble_chart_components(
                plot_table,
                metrics,
                ref_group,
                global_scales,
                selection,
                fairness_threshold,
                chart_height,
                chart_sizes["disparity_chart_width"],
                accessibility_mode,
                concat_chart=True,
            )
            + disparity_title
        )
        .resolve_scale(y="independent", size="independent")
        .properties(height=chart_height, width=chart_sizes["disparity_chart_width"])
    )

    # METRIC CHART
    metric_chart = (
        (
            get_metric_bubble_chart_components(
                plot_table,
                metrics,
                ref_group,
                global_scales,
                selection,
                fairness_threshold,
                chart_height,
                chart_sizes["metric_chart_width"],
                accessibility_mode,
                concat_chart=True,
            )
            + metric_title
            + draw_legend(global_scales, selection, chart_sizes["metric_chart_width"])
        )
        .resolve_scale(y="independent")
        .properties(height=chart_height, width=chart_sizes["metric_chart_width"])
    )

    full_chart = (
        alt.hconcat(disparity_chart, metric_chart, bounds="flush", spacing=0)
        .configure_view(strokeWidth=0)
        .configure_axisLeft(
            labelFontSize=Metric_Axis.label_font_size,
            labelColor=Metric_Axis.label_color,
            labelFont=FONT,
        )
        .configure_title(
            align="center",
            baseline="middle",
            font=FONT,
            fontWeight=Chart_Title.font_weight,
            fontSize=Chart_Title.font_size,
            color=Chart_Title.font_color,
            dx=chart_width / 2,
        )
        .properties(
            title=f"{attribute.title()}", padding=Concat_Chart.full_chart_padding
        )
    )

    return full_chart



import pandas as pd
import altair as alt
import math

from aequitas.plot.commons.style.classes import Legend
from aequitas.plot.commons.style.text import FONT
from aequitas.plot.commons.helpers import no_axis


DUMMY_DF = pd.DataFrame({"a": [1, 1], "b": [0, 0]})


def draw_legend(global_scales, selection, chart_width):
    """Draws the interactive group's colors legend for the chart."""

    groups = global_scales["color"].domain
    labels = groups.copy()
    labels[0] = labels[0] + " [REF]"
    legend_df = pd.DataFrame({"attribute_value": groups, "label": labels})

    # Position the legend to the right of the chart

    title_text_x_position = chart_width
    title_text_height = Legend.title_font_size + Legend.title_margin_bottom
    subtitle_text_height = Legend.font_size + Legend.vertical_spacing

    entries_circles_x_position = title_text_x_position + Legend.horizontal_spacing
    entries_text_x_position = (
        title_text_x_position + 2 * Legend.circle_radius + Legend.horizontal_spacing
    )

    # Title of the legend.
    title_text = (
        alt.Chart(DUMMY_DF)
        .mark_text(
            align="left",
            baseline="middle",
            color=Legend.font_color,
            fontSize=Legend.title_font_size,
            font=FONT,
            fontWeight=Legend.title_font_weight,
        )
        .encode(
            x=alt.value(title_text_x_position),
            y=alt.value(Legend.margin_top),
            text=alt.value("Groups"),
        )
    )

    # Subtitle text that explains how to interact with the legend.
    subtitle_text = (
        alt.Chart(DUMMY_DF)
        .mark_text(
            align="left",
            baseline="middle",
            color=Legend.font_color,
            fontSize=Legend.font_size,
            font=FONT,
            fontWeight=Legend.font_weight,
        )
        .encode(
            x=alt.value(title_text_x_position),
            y=alt.value(Legend.margin_top + title_text_height),
            text=alt.value("Click to highlight a group."),
        )
    )

    # Conditionally color each legend item
    # If the group is selected, it is colored according to the group color scale, otherwise it is faded
    color_encoding = alt.condition(
        selection,
        alt.Color("attribute_value:N", scale=global_scales["color"], legend=None),
        alt.value(Legend.color_faded),
    )

    # Offset the positioning of the legend items after the subitlr text
    legend_start_y_position = (
        Legend.margin_top + title_text_height + subtitle_text_height
    )

    y_scale = alt.Scale(
        domain=groups,
        range=[
            legend_start_y_position,
            legend_start_y_position
            # number of legend elements x text size
            + (len(groups) * Legend.font_size)
            # (number of "spacings" + start and end "spacings") x spacing
            + ((len(groups) + 1) * Legend.vertical_spacing),
        ],
    )

    # Calculate circle size from radius
    entries_circle_size = Legend.circle_radius * math.pi ** 2

    # Draw color squares for each group
    entries_circles = (
        alt.Chart(legend_df)
        .mark_point(filled=True, opacity=1, size=entries_circle_size)
        .encode(
            x=alt.value(entries_circles_x_position),
            y=alt.Y("attribute_value:N", scale=y_scale, axis=no_axis()),
            color=color_encoding,
            shape=alt.Shape(
                "attribute_value:N", scale=global_scales["shape"], legend=None
            ),
        )
        .add_selection(selection)
    )
    trigger_text = alt.selection_multi(empty="all", fields=["attribute_value"])

    # Draw colored label for each group
    entries_text = (
        alt.Chart(legend_df)
        .mark_text(
            align="left",
            baseline="middle",
            font=FONT,
            fontSize=Legend.font_size,
            fontWeight=Legend.font_weight,
        )
        .encode(
            x=alt.value(entries_text_x_position),
            y=alt.Y("attribute_value:N", scale=y_scale, axis=no_axis()),
            text=alt.Text("label:N"),
            color=color_encoding,
        )
        .add_selection(trigger_text)
    )

    return entries_circles + entries_text + subtitle_text + title_text



import math
import altair as alt
import pandas as pd

from aequitas.plot.commons.helpers import (
    no_axis,
    transform_ratio,
    calculate_chart_size_from_elements,
    to_list,
    format_number,
)
from aequitas.plot.commons.tooltips import (
    get_tooltip_text_group_size,
    get_tooltip_text_disparity_explanation,
    get_tooltip_text_parity_test_explanation,
)
from aequitas.plot.commons.style.classes import (
    Title,
    Subtitle,
    Parity_Result,
    Annotation,
    Legend,
)
from aequitas.plot.commons.style.text import FONT
from aequitas.plot.commons.style.sizes import Summary_Chart
from aequitas.plot.commons import initializers as Initializer
from aequitas.plot.commons import validators as Validator

# Altair 2.4.1 requires that all chart receive a dataframe, for charts that don't need it
# (like most annotations), we pass the following dummy dataframe to reduce the complexity of the resulting vega spec.
DUMMY_DF = pd.DataFrame({"a": [1, 1], "b": [0, 0]})


def __get_scales(max_num_groups):
    """Creates an Altair scale for the color of the parity test result, and another
    for the x axis of the group circles subplot."""

    scales = dict()

    # COLOR
    scales["color"] = alt.Scale(
        domain=["Reference", "Pass", "Fail"], range=Parity_Result.color_palette
    )

    # GROUP CIRCLES X AXIS
    scales["circles_x"] = alt.Scale(
        domain=[-max_num_groups, max_num_groups], nice=False
    )

    return scales


def __get_size_constants(
    chart_height, chart_width, num_attributes, num_metrics, max_num_groups
):
    """Calculates the heights, widths and spacings of the components of the summary chart
    based on the provided desired overall chart height and width, as well as the number of
    attributes (columns) and metrics (lines)."""

    size_constants = dict(
        # Chart sizes
        attribute_titles_height=0.15 * chart_height,
        line_spacing=0.2 * chart_height / num_metrics,
        line_height=Summary_Chart.line_height_ratio * chart_height / num_metrics,
        metric_titles_width=0.1 * chart_width,
        column_spacing=0.15 * chart_width / num_attributes,
        column_width=Summary_Chart.column_width_ratio * chart_width / num_attributes,
        # Circle size
        ## Conditional definition of the size where for each additional unit in
        ## max_num_groups, we subtract 25 squared pixels from the area of the
        ## circle, which has the base value of 350 for 0 groups. From max_num_groups
        ## equal to 13 or more, we keep the size at the minimum value of 25 to
        ## make sure the circles are visible.
        group_circle_size=-25 * max_num_groups + 350 if max_num_groups < 13 else 25,
    )
    return size_constants


def __draw_attribute_title(attribute, width, size_constants):
    """Draws a single attribute's title."""

    return (
        alt.Chart(DUMMY_DF)
        .mark_text(
            align="center",
            baseline="middle",
            font=FONT,
            size=Title.font_size,
            color=Title.font_color,
            fontWeight=Title.font_weight,
        )
        .encode(
            text=alt.value(attribute.title()),
        )
        .properties(width=width, height=size_constants["attribute_titles_height"])
    )


def __draw_metric_line_titles(metrics, size_constants):
    """Draws left hand side titles for metrics."""

    metric_line_titles = []

    for metric in metrics:
        # METRIC TITLE
        metric_title = (
            alt.Chart(DUMMY_DF)
            .transform_calculate(y_position="1.2")
            .mark_text(
                align="center",
                baseline="middle",
                font=FONT,
                fontWeight=Title.font_weight,
                size=Title.font_size,
                color=Title.font_color,
            )
            .encode(
                alt.Y("y_position:Q", scale=alt.Scale(domain=[3, 1]), axis=no_axis()),
                text=alt.value(metric.upper()),
            )
        )

        # GROUPS TEXT
        group_circles_title = (
            alt.Chart(DUMMY_DF)
            .transform_calculate(y_position="2")
            .mark_text(
                align="center",
                baseline="middle",
                font=FONT,
                size=Subtitle.font_size,
                color=Subtitle.font_color,
            )
            .encode(
                alt.Y("y_position:Q", scale=alt.Scale(domain=[3, 1]), axis=no_axis()),
                text=alt.value("Groups"),
            )
        )

        # PERCENT. POP TEXT
        population_percentage_title = (
            alt.Chart(DUMMY_DF)
            .transform_calculate(y_position="2.7")
            .mark_text(
                align="center",
                baseline="middle",
                font=FONT,
                size=Subtitle.font_size,
                color=Subtitle.font_color,
            )
            .encode(
                alt.Y("y_position:Q", scale=alt.Scale(domain=[3, 1]), axis=no_axis()),
                text=alt.value("% Pop."),
            )
        )

        metric_line_titles.append(
            (
                metric_title + group_circles_title + population_percentage_title
            ).properties(
                height=size_constants["line_height"],
                width=size_constants["metric_titles_width"],
            )
        )

    # EMPTY CORNER SPACE
    # To make sure that the attribute columns align properly with the title column, we need to create a blank
    # space of the same size of the attribute titles. For this purpose, we use the same function (__draw_attribute_title)
    # and pass in an empty string so that nothing is actually drawn.
    top_left_corner_space = __draw_attribute_title(
        "", size_constants["metric_titles_width"], size_constants
    )

    # CONCATENATE SUBPLOTS
    metric_titles = alt.vconcat(
        top_left_corner_space,
        *metric_line_titles,
        spacing=size_constants["line_spacing"],
        bounds="flush",
    )

    return metric_titles


def __get_parity_result_variable(row, metric, fairness_threshold):
    """ Creates parity test result variable for each provided row, separating the Reference group from the passing ones."""
    if row["attribute_value"] == row["ref_group_value"]:
        return "Reference"
    elif abs(row[f"{metric}_disparity_scaled"]) < fairness_threshold - 1:
        return "Pass"
    else:
        return "Fail"


def __draw_parity_result_text(parity_result, color_scale):
    """Draws the uppercased text result of the provided parity test (Pass, Fail or Reference),
    color-coded according to the provided Altair scale."""

    return (
        alt.Chart(pd.DataFrame({"parity_result": parity_result}, index=[0]))
        .transform_calculate(y_position="1")
        .mark_text(
            align="center",
            baseline="middle",
            font=FONT,
            size=Parity_Result.font_size,
            fontWeight=Parity_Result.font_weight,
        )
        .encode(
            alt.Y("y_position:Q", scale=alt.Scale(domain=[3, 1]), axis=no_axis()),
            alt.Color(
                "parity_result:O", scale=color_scale, legend=alt.Legend(title="")
            ),
            text=alt.value(parity_result.upper()),
        )
    )


def __draw_population_bar(population_bar_df, metric, color_scale):
    """ Draws a stacked bar of the sum of the percentage of population of the groups that obtained each result for the parity test."""
    population_bar_tooltips = [
        alt.Tooltip(field=f"{metric}_parity_result", type="nominal", title="Parity"),
        alt.Tooltip(
            field="tooltip_group_size",
            type="nominal",
            title="Size",
        ),
        alt.Tooltip(field="tooltip_groups_name_size", type="nominal", title="Groups"),
    ]

    population_bar = (
        alt.Chart(population_bar_df)
        .transform_calculate(y_position="2.8")
        .mark_bar(size=6, stroke="white")
        .encode(
            alt.X("sum(group_size):Q", stack="normalize", axis=no_axis()),
            alt.Y("y_position:Q", scale=alt.Scale(domain=[3, 1]), axis=no_axis()),
            alt.Color(
                f"{metric}_parity_result:O",
                scale=color_scale,
                legend=alt.Legend(
                    title="Parity Test",
                    padding=20,
                ),
            ),
            tooltip=population_bar_tooltips,
        )
    )

    return population_bar


def __draw_group_circles(plot_df, metric, scales, size_constants):
    """Draws a circle for each group, color-coded by the result of the parity test.
    The groups are spread around the central reference group according to their disparity."""

    circle_tooltip_encoding = [
        alt.Tooltip(field="attribute_value", type="nominal", title="Group"),
        alt.Tooltip(field="tooltip_group_size", type="nominal", title="Group Size"),
        alt.Tooltip(
            field=f"tooltip_parity_test_explanation_{metric}",
            type="nominal",
            title="Parity Test",
        ),
        alt.Tooltip(
            field=f"tooltip_disparity_explanation_{metric}",
            type="nominal",
            title="Disparity",
        ),
        alt.Tooltip(
            field=f"{metric}",
            type="quantitative",
            format=".2f",
            title=f"{metric}".upper(),
        ),
    ]

    return (
        alt.Chart(plot_df)
        .transform_calculate(y_position="2")
        .mark_circle(opacity=1)
        .encode(
            alt.X(
                f"{metric}_disparity_rank:Q", scale=scales["circles_x"], axis=no_axis()
            ),
            alt.Y("y_position:Q", scale=alt.Scale(domain=[3, 1]), axis=no_axis()),
            alt.Color(
                f"{metric}_parity_result:O",
                scale=scales["color"],
                legend=alt.Legend(title=""),
            ),
            size=alt.value(size_constants["group_circle_size"]),
            tooltip=circle_tooltip_encoding,
        )
    )


def __draw_parity_test_explanation(fairness_threshold, x_position):
    """Draw text that explains what does pass/fail mean in the parity test results."""

    explanation_text = alt.Chart(DUMMY_DF).mark_text(
        baseline="top",
        align="left",
        font=FONT,
        fill=Annotation.font_color,
        fontSize=Annotation.font_size,
        fontWeight=Annotation.font_weight,
    )

    explanation_text_group = explanation_text.encode(
        x=alt.value(x_position),
        y=alt.value(0),
        text=alt.value(
            f"For a group to pass the parity test its disparity to the reference group cannot exceed the fairness threshold ({fairness_threshold})."
        ),
    )

    explanation_text_attribute = explanation_text.encode(
        x=alt.value(x_position),
        y=alt.value(Annotation.font_size * Annotation.line_spacing),
        text=alt.value(
            f"An attribute passes the parity test for a given metric if all its groups pass the test."
        ),
    )

    return explanation_text_group + explanation_text_attribute


def __create_population_bar_df(attribute_df, metric):
    """Creates a pandas aggregation of the attribute_df by parity result, along with the
    list of groups tooltip variable."""

    attribute_df["group_size_formatted"] = attribute_df.apply(
        lambda row: format_number(row["group_size"]), axis=1
    )
    attribute_df["tooltip_groups_name_size"] = (
        attribute_df["attribute_value"]
        + " ("
        + attribute_df["group_size_formatted"].astype(str)
        + ")"
    )
    population_bar_df = (
        attribute_df.groupby(by=f"{metric}_parity_result")
        .agg(
            {
                "attribute_name": min,
                "total_entities": min,
                "group_size": sum,
                "tooltip_groups_name_size": lambda x: ", ".join(x),
            }
        )
        .reset_index()
    )

    population_bar_df["tooltip_group_size"] = population_bar_df.apply(
        lambda row: get_tooltip_text_group_size(
            row["group_size"], row["total_entities"]
        ),
        axis=1,
    )

    return population_bar_df


def __create_group_rank_variable(attribute_df, metric):
    """ Creates the disparity rank variable for the given metric, centered around 0 (the Reference Group's value). """

    # RANK
    attribute_df[f"{metric}_disparity_rank"] = attribute_df[
        f"{metric}_disparity_scaled"
    ].rank(method="first")

    # REFERENCE GROUP RANK
    reference_rank = attribute_df.loc[
        attribute_df[f"{metric}_parity_result"] == "Reference"
    ][f"{metric}_disparity_rank"].iloc[0]

    # CENTERED RANK
    attribute_df[f"{metric}_disparity_rank"] = (
        attribute_df[f"{metric}_disparity_rank"] - reference_rank
    )


def __create_tooltip_variables(attribute_df, metric, fairness_threshold):
    """ Creates disparity explanation and formatted group size tooltip variables. """

    # PARITY TEST EXPLANATION
    attribute_df[f"tooltip_parity_test_explanation_{metric}"] = attribute_df.apply(
        lambda row: get_tooltip_text_parity_test_explanation(
            row[f"{metric}_parity_result"],
            metric,
            fairness_threshold,
        ),
        axis=1,
    )

    # DISPARITY EXPLANATION
    ref_group = attribute_df["ref_group_value"].iloc[0]

    attribute_df[f"tooltip_disparity_explanation_{metric}"] = attribute_df.apply(
        lambda row: get_tooltip_text_disparity_explanation(
            row[f"{metric}_disparity_scaled"],
            row["attribute_value"],
            metric,
            ref_group,
        ),
        axis=1,
    )
    # FORMATTED GROUP SIZE

    attribute_df["tooltip_group_size"] = attribute_df.apply(
        lambda row: get_tooltip_text_group_size(
            row["group_size"], row["total_entities"]
        ),
        axis=1,
    )


def __create_disparity_variables(attribute_df, metric, fairness_threshold):
    """ Creates scaled disparity, parity test result & disparity explanation tooltip variables. """

    # SCALED DISPARITY VALUE
    attribute_df[f"{metric}_disparity_scaled"] = attribute_df.apply(
        lambda row: transform_ratio(row[f"{metric}_disparity"]), axis=1
    )

    # PARITY RESULT
    attribute_df[f"{metric}_parity_result"] = attribute_df.apply(
        __get_parity_result_variable,
        metric=metric,
        fairness_threshold=fairness_threshold,
        axis=1,
    )


def __get_attribute_column(
    attribute_df, metrics, scales, attribute, size_constants, fairness_threshold
):
    """ Returns a vertical concatenation of all elements of all metrics for each attribute's column."""

    metric_summary = []

    for metric in metrics:
        # CREATE VARIABLES IN DF
        __create_disparity_variables(attribute_df, metric, fairness_threshold)
        __create_tooltip_variables(attribute_df, metric, fairness_threshold)
        __create_group_rank_variable(attribute_df, metric)

        # PARITY RESULT TEXT
        ## The parity result is equal to the "worst" of each group's results
        ## If one group fails the parity test, the whole metric fails (for that attribute)
        parity_result = attribute_df.loc[
            attribute_df[f"{metric}_parity_result"] != "Reference"
        ][f"{metric}_parity_result"].min()

        parity_result_text = __draw_parity_result_text(parity_result, scales["color"])

        # GROUP CIRCLES
        group_circles = __draw_group_circles(
            attribute_df, metric, scales, size_constants
        )

        # POPULATION BAR
        population_bar_df = __create_population_bar_df(attribute_df, metric)
        population_bar = __draw_population_bar(
            population_bar_df, metric, scales["color"]
        )

        # LAYERING
        metric_summary.append(
            (parity_result_text + population_bar + group_circles)
            .properties(
                width=size_constants["column_width"],
                height=size_constants["line_height"],
            )
            .resolve_scale(x="independent")
        )

    # ATTRIBUTE TITLE
    attribute_title = __draw_attribute_title(
        attribute, size_constants["column_width"], size_constants
    )

    return alt.vconcat(
        attribute_title,
        *metric_summary,
        bounds="flush",
        spacing=size_constants["line_spacing"],
    )


def plot_summary_chart(
    disparity_df,
    metrics_list,
    attributes_list=None,
    fairness_threshold=1.25,
    chart_height=None,
    chart_width=None,
):
    """Draws chart that summarizes the parity results for the provided metrics across the existing attributes.
    This includes an overall result, the specific results by each attribute's groups as well as the percentage
    of population by result.

    :param disparity_df: a dataframe generated by the Aequitas Bias class
    :type disparity_df: pandas.core.frame.DataFrame
    :param metrics_list: a list of the metrics of interest
    :type metrics_list: list
    :param attributes_list: a list of the attributes of interest, defaults to using all in the dataframe
    :type attributes_list: list, optional
    :param fairness_threshold: a value for the maximum allowed disparity, defaults to 1.25
    :type fairness_threshold: float, optional
    :param chart_height: a value (in pixels) for the height of the chart
    :type chart_height: int, optional
    :param chart_width: a value (in pixels) for the width of the chart
    :type chart_width: int, optional

    :return: the full summary chart
    :rtype: Altair chart object
    """

    ## If a specific list of attributes was not passed, use all from df
    (
        metrics,
        attributes,
        chart_height,
        chart_width,
    ) = Initializer.prepare_summary_chart(
        disparity_df,
        metrics_list,
        attributes_list,
        fairness_threshold,
        chart_height,
        chart_width,
        Summary_Chart,
    )

    num_metrics = len(metrics)
    num_attributes = len(attributes)

    max_num_groups = max(
        disparity_df.loc[disparity_df["attribute_name"].isin(attributes)]
        .groupby(by="attribute_name")["attribute_value"]
        .count()
    )

    size_constants = __get_size_constants(
        chart_height, chart_width, num_attributes, num_metrics, max_num_groups
    )

    Validator.chart_size_summary(size_constants, num_metrics, num_attributes)

    # SCALES
    scales = __get_scales(max_num_groups)

    # METRIC TITLES
    metric_titles = __draw_metric_line_titles(metrics, size_constants)

    # RELEVANT FIELDS
    viz_fields = [
        "attribute_name",
        "attribute_value",
        "group_size",
        "total_entities",
        f"{metrics[0]}_ref_group_value",
        *metrics,
    ]
    viz_fields += [f"{metric}_disparity" for metric in metrics]

    attribute_columns = []

    for attribute in attributes:
        # CREATE ATTRIBUTE DF
        attribute_df = (
            disparity_df[viz_fields]
            .loc[disparity_df["attribute_name"] == attribute]
            .copy(deep=True)
        )

        attribute_df.rename(
            columns={f"{metrics[0]}_ref_group_value": "ref_group_value"}, inplace=True
        )

        # ATTRIBUTE COLUMN
        attribute_column = __get_attribute_column(
            attribute_df,
            metrics,
            scales,
            attribute,
            size_constants,
            fairness_threshold,
        )

        attribute_columns.append((attribute_column))

    # CONCATENATE ATTRIBUTE COLUMNS
    summary_chart_columns = alt.hconcat(
        *attribute_columns,
        bounds="flush",
        spacing=size_constants["column_spacing"] + size_constants["column_width"],
    )
    # ADD METRIC TITLES
    summary_chart_table = alt.hconcat(
        metric_titles,
        summary_chart_columns,
        bounds="flush",
        spacing=size_constants["metric_titles_width"]
        + size_constants["column_spacing"],
    )

    summary_chart_explanation = __draw_parity_test_explanation(
        fairness_threshold, size_constants["column_spacing"] / 2
    )

    full_summary_chart = (
        alt.vconcat(summary_chart_table, summary_chart_explanation)
        .properties(padding=Summary_Chart.full_chart_padding)
        .configure_legend(
            labelFont=FONT,
            labelColor=Legend.font_color,
            labelFontSize=Legend.font_size,
            titleFont=FONT,
            titleColor=Legend.font_color,
            titleFontSize=Legend.title_font_size,
            titleFontWeight=Legend.title_font_weight,
            titlePadding=Legend.title_margin_bottom + Legend.vertical_spacing,
        )
        .configure_view(strokeWidth=0)
    )

    return full_summary_chart



from aequitas.plot.summary_chart import plot_summary_chart as summary
from aequitas.plot.bubble_disparity_chart import plot_disparity_bubble_chart as disparity
from aequitas.plot.bubble_metric_chart import plot_metric_bubble_chart as absolute
from aequitas.plot.bubble_concatenation_chart import plot_concatenated_bubble_charts as disparity_absolute
from aequitas.plot.xy_metrics_chart import plot_xy_metrics_chart as scatter


import altair as alt

from aequitas.plot.commons import validators as Validator
from aequitas.plot.commons import scales as Scales
from aequitas.plot.commons.helpers import to_list, calculate_chart_size_from_elements


def __sanitize_metrics(metrics_list):
    try:
        return [metric.lower() for metric in to_list(metrics_list)]
    except AttributeError:
        print('"metrics_list" must be a list of strings.')


def __filter_df(disparity_df, metrics, attribute, xy_plot=False):
    viz_fields = [
        "attribute_name",
        "attribute_value",
        "group_size",
        "total_entities",
    ]
    viz_fields += metrics

    if not xy_plot:
        viz_fields += [f"{metric}_disparity" for metric in metrics]

    plot_table = disparity_df[viz_fields][
        disparity_df["attribute_name"] == attribute
    ].copy(deep=True)

    return plot_table


def __validate_data_inputs(disparity_df, metrics, attributes, fairness_threshold):
    Validator.metrics(metrics)
    Validator.dataframe(disparity_df, metrics)
    Validator.attributes(disparity_df, attributes)
    Validator.fairness_threshold(fairness_threshold)


def __calculate_default_sizes(
    chart_height, chart_width, chart_default_sizes, metrics=None, attributes=None
):
    if chart_height is None:
        num_metrics = len(metrics)

        chart_height = calculate_chart_size_from_elements(
            chart_default_sizes.vertical_header_height,
            chart_default_sizes.line_height,
            num_metrics,
        )

    if chart_width is None:
        num_attributes = len(attributes)

        chart_width = calculate_chart_size_from_elements(
            chart_default_sizes.horizontal_header_width,
            chart_default_sizes.column_width,
            num_attributes,
        )

    return (chart_height, chart_width)


def __get_global_scales(
    plot_table, ref_group, metrics, chart_height, accessibility_mode
):
    global_scales = dict()
    global_scales["color"] = Scales.get_color_scale(plot_table, ref_group)
    global_scales["bubble_size"] = Scales.get_bubble_size_scale(
        plot_table, metrics, chart_height
    )
    global_scales["shape"] = Scales.get_shape_scale(
        plot_table, ref_group, accessibility_mode
    )

    return global_scales


def prepare_bubble_chart(
    disparity_df,
    metrics_list,
    attribute,
    fairness_threshold,
    chart_height,
    chart_width,
    chart_default_sizes,
    accessibility_mode,
):
    metrics = __sanitize_metrics(metrics_list)

    __validate_data_inputs(disparity_df, metrics, attribute, fairness_threshold)

    ## Calculate default chart sizes
    num_metrics = len(metrics)

    (chart_height, chart_width) = __calculate_default_sizes(
        chart_height, chart_width, chart_default_sizes, metrics
    )

    Validator.chart_size_bubble(
        chart_width, chart_height, chart_default_sizes, num_metrics
    )

    # GET REF GROUP
    ref_group = disparity_df.loc[disparity_df["attribute_name"] == attribute][
        f"{metrics[0]}_ref_group_value"
    ].iloc[0]

    plot_table = __filter_df(disparity_df, metrics, attribute)

    # GLOBAL CHART MECHANICS
    ## SCALES: COLOR, BUBBLE SIZE, SHAPE
    global_scales = __get_global_scales(
        plot_table, ref_group, metrics, chart_height, accessibility_mode
    )

    # SELECTION
    selection = alt.selection_multi(empty="all", fields=["attribute_value"])

    return (
        plot_table,
        metrics,
        ref_group,
        global_scales,
        chart_height,
        chart_width,
        selection,
    )


def prepare_summary_chart(
    disparity_df,
    metrics_list,
    attributes_list,
    fairness_threshold,
    chart_height,
    chart_width,
    chart_default_sizes,
):
    ## If a specific list of attributes was not passed, use all from df
    if fairness_threshold is None:
        raise ValueError("Fairness threshold cannot be None for the Summary chart.")

    metrics = __sanitize_metrics(metrics_list)

    if attributes_list is None:
        attributes = list(disparity_df["attribute_name"].unique())
    else:
        attributes = to_list(attributes_list)

    __validate_data_inputs(disparity_df, metrics, attributes, fairness_threshold)

    # CHART SIZES
    (chart_height, chart_width) = __calculate_default_sizes(
        chart_height, chart_width, chart_default_sizes, metrics, attributes
    )

    return (metrics, attributes, chart_height, chart_width)


def prepare_xy_chart(
    disparity_df,
    x_metric,
    y_metric,
    attribute,
    fairness_threshold,
    chart_height,
    chart_width,
    chart_default_sizes,
    accessibility_mode,
):
    metrics = __sanitize_metrics([x_metric, y_metric])

    x_metric = metrics[0]
    y_metric = metrics[1]

    __validate_data_inputs(disparity_df, metrics, attribute, fairness_threshold)

    if (chart_width is None) and (chart_height is None):
        chart_width = chart_default_sizes.full_side
        chart_height = chart_default_sizes.full_side
    elif chart_width is None:
        chart_width = chart_height
    elif chart_height is None:
        chart_height = chart_width

    Validator.chart_size_xy(chart_width, chart_height)

    # REF GROUP DEFINITION
    ## ref_group is taken from x_metric as long as it's not part of the metrics that
    ## do not have disparity variables (such as prev)
    if x_metric in Validator.NON_DISPARITY_METRICS_LIST:
        ref_group = disparity_df[disparity_df["attribute_name"] == attribute].iloc[0][
            f"{y_metric}_ref_group_value"
        ]
    else:
        ref_group = disparity_df[disparity_df["attribute_name"] == attribute].iloc[0][
            f"{x_metric}_ref_group_value"
        ]

    plot_table = __filter_df(disparity_df, metrics, attribute, xy_plot=True)

    # SCALES
    global_scales = __get_global_scales(
        plot_table, ref_group, metrics, chart_height, accessibility_mode
    )

    # SELECTION
    interactive_selection_group = alt.selection_multi(
        empty="all", fields=["attribute_value"]
    )

    return (
        plot_table,
        x_metric,
        y_metric,
        ref_group,
        global_scales,
        chart_height,
        chart_width,
        interactive_selection_group,
    )



# COLORS
GRAY = "rgb(117,117,117)"
LIGHT_GRAY = "rgb(224,224,224)"
RED = "rgb(217, 46, 28)"
TEAL = "rgb(20, 152, 181)"

BLUE = "rgb(100, 143, 255)"
ORANGE = "rgb(254, 97, 0)"
PINK = "rgb(220, 38, 127)"
YELLOW = "rgb(255, 176, 0)"
PURPLE = "rgb(120, 94, 240)"

# NAMED COLORS
REFERENCE = GRAY
FADED = LIGHT_GRAY
THRESHOLD = RED
PARITY_PASS = TEAL
PARITY_FAIL = THRESHOLD
PARITY_RESULT_PALETTE = [PARITY_PASS, PARITY_FAIL]


# Colors will be selected for the groups in the chart by the
# order that they are in this palette array. Therefore, the
# order should maximamize constrast for smaller subsets [1:n].
CATEGORICAL_PALETTE_COLOR_SAFE = [
    BLUE,
    ORANGE,
    PINK,
    YELLOW,
    PURPLE,
]

CATEGORICAL_PALETTE_ALTERNATIVE = [
    "rgb(106, 172, 250)",
    "rgb(62, 204, 156)",
    "rgb(255, 163, 31)",
    "rgb(255, 109, 135)",
    "rgb(182, 135, 231)",
]

CATEGORICAL_COLOR_PALETTE = CATEGORICAL_PALETTE_COLOR_SAFE



import math
import altair as alt
import pandas as pd

from aequitas.plot.commons.helpers import no_axis, transform_ratio
from aequitas.plot.commons.legend import draw_legend
from aequitas.plot.commons.scales import get_chart_size_range
from aequitas.plot.commons.tooltips import (
    get_tooltip_text_group_size,
    get_tooltip_text_disparity_explanation,
)
from aequitas.plot.commons.style.classes import (
    Metric_Axis,
    Axis,
    Threshold_Band,
    Threshold_Rule,
    Annotation,
    Bubble,
    Rule,
    Chart_Title,
)
from aequitas.plot.commons.style.sizes import Metric_Chart
from aequitas.plot.commons.style.text import FONT
from aequitas.plot.commons import initializers as Initializer

# Altair 2.4.1 requires that all chart receive a dataframe, for charts that don't need it
# (like most annotations), we pass the following dummy dataframe to reduce the complexity of the resulting vega spec.
DUMMY_DF = pd.DataFrame({"a": [1, 1], "b": [0, 0]})


def __get_position_scales(metrics, chart_height, chart_width, concat_chart):
    """Computes the scales for x and y encodings to be used in the metric bubble chart."""

    position_scales = dict()

    # METRICS VALUES SCALE
    x_range = get_chart_size_range(chart_width, Metric_Chart.padding_x)
    x_domain = [0, 1]
    position_scales["x"] = alt.Scale(domain=x_domain, range=x_range)

    # METRICS LABELS SCALE
    y_range = get_chart_size_range(chart_height, Metric_Chart.padding_y)
    y_domain = [metric.upper() for metric in metrics]
    position_scales["y"] = alt.Scale(domain=y_domain, range=y_range)

    return position_scales


def __draw_metrics_rules(metrics, scales, concat_chart):
    """Draws an horizontal rule for each metric where the bubbles will be positioned."""

    metrics_labels = [metric.upper() for metric in metrics]

    if concat_chart:
        y_axis = no_axis()
    else:
        y_axis = alt.Axis(
            domain=False,
            ticks=False,
            orient="left",
            labelAngle=Metric_Axis.label_angle,
            labelPadding=Metric_Axis.label_padding,
            title="",
        )

    horizontal_rules = (
        alt.Chart(pd.DataFrame({"y_position": metrics_labels, "start": 0, "end": 1}))
        .mark_rule(
            strokeWidth=Metric_Axis.stroke_width,
            stroke=Metric_Axis.stroke,
            tooltip="",
        )
        .encode(
            y=alt.Y("y_position:N", scale=scales["y"], axis=y_axis),
            x=alt.X("start:Q", scale=scales["x"]),
            x2="end:Q",
        )
    )

    return horizontal_rules


def __draw_domain_rules(scales):
    """Draws vertical lines at the ends of each metric's domain."""

    x_domain = [0, 1]
    domain_rules_df = pd.DataFrame({"x_position": x_domain})

    vertical_domain_rules = (
        alt.Chart(domain_rules_df)
        .mark_rule(
            strokeWidth=Rule.stroke_width,
            stroke=Rule.stroke,
            tooltip="",
        )
        .encode(
            x=alt.X("x_position:Q", scale=scales["x"]),
            y=alt.value(scales["y"]["range"][0]),
            y2=alt.value(scales["y"]["range"][1]),
        )
    )

    return vertical_domain_rules


def __draw_x_ticks_labels(scales, chart_height):
    """Draws the numbers in the horizontal axis."""

    axis_values = [0, 0.25, 0.5, 0.75, 1]

    axis_df = pd.DataFrame({"value": axis_values})

    x_ticks_labels = (
        alt.Chart(axis_df)
        .mark_text(
            tooltip="",
            align="center",
            fontSize=Axis.label_font_size,
            font=FONT,
            fontWeight=Axis.label_font_weight,
            color=Axis.label_color,
        )
        .encode(
            text=alt.Text("value:N"),
            x=alt.X(
                "value:Q",
                scale=scales["x"],
            ),
            y=alt.value(Metric_Chart.padding_y * chart_height * 0.7),
        )
    )

    return x_ticks_labels


def __draw_threshold_rules(threshold_df, scales, position, accessibility_mode=False):
    """Draws fairness threshold rules: red lines that mark the defined fairness threshold in the chart."""
    stroke_color = (
        Threshold_Rule.stroke_accessible
        if accessibility_mode
        else Threshold_Rule.stroke
    )

    threshold_rules = (
        alt.Chart(threshold_df)
        .mark_rect(
            stroke=stroke_color,
            opacity=Threshold_Rule.opacity,
            strokeWidth=Threshold_Rule.stroke_width,
            tooltip="",
        )
        .encode(
            x=alt.X(
                field=f"{position}_threshold_value",
                type="quantitative",
                scale=scales["x"],
            ),
            x2=f"{position}_threshold_value:Q",
            y=alt.Y(field="metric", type="nominal", scale=scales["y"], axis=no_axis()),
        )
    )
    return threshold_rules


def __draw_threshold_bands(threshold_df, scales, accessibility_mode=False):
    """Draws fairness threshold bands: regions painted red where the metric value is above the defined fairness threshold."""

    fill_color = (
        Threshold_Band.color_accessible if accessibility_mode else Threshold_Band.color
    )

    lower_threshold_band = (
        alt.Chart(threshold_df)
        .mark_rect(fill=fill_color, opacity=Threshold_Band.opacity, tooltip="")
        .encode(
            y=alt.Y(field="metric", type="nominal", scale=scales["y"], axis=no_axis()),
            x=alt.X("lower_threshold_value:Q", scale=scales["x"]),
            x2="lower_end:Q",
        )
    )

    upper_threshold_band = (
        alt.Chart(threshold_df)
        .mark_rect(fill=fill_color, opacity=Threshold_Band.opacity, tooltip="")
        .encode(
            y=alt.Y(field="metric", type="nominal", scale=scales["y"], axis=no_axis()),
            x=alt.X("upper_threshold_value:Q", scale=scales["x"]),
            x2="upper_end:Q"
            # scales["x"]["range"][1]
            # Replicating the approach of the lower_threshold_band doesn't work...
        )
    )

    return lower_threshold_band + upper_threshold_band


def __draw_threshold_text(
    fairness_threshold, ref_group, chart_height, accessibility_mode=False
):
    """Draw text that helps to read the fairness threshold elements of the chart."""
    font_color = (
        Annotation.font_color if accessibility_mode else Annotation.font_color_threshold
    )

    threshold_text = (
        alt.Chart(DUMMY_DF)
        .mark_text(
            baseline="top",
            align="left",
            font=FONT,
            fill=font_color,
            fontSize=Annotation.font_size,
            fontWeight=Annotation.font_weight,
            tooltip="",
        )
        .encode(
            x=alt.value(0),
            y=alt.value(chart_height * (1 - 2 / 3 * Metric_Chart.padding_y)),
            text=alt.value(
                f"The metric value for any group should not be {fairness_threshold} (or more) times smaller or larger than that of the reference group {ref_group}."
            ),
        )
    )

    return threshold_text


def __get_threshold_elements(
    plot_table,
    metrics,
    fairness_threshold,
    ref_group,
    scales,
    chart_height,
    concat_chart,
    accessibility_mode=False,
):
    """Gets fairness threshold elements (rules, bands and text) for the chart."""

    lower_values = []
    upper_values = []

    # CREATE THRESHOLD DF
    for metric in metrics:
        # Get reference value for each metric
        ref_group_value = plot_table.loc[plot_table["attribute_value"] == ref_group][
            metric
        ].iloc[0]

        # Enforce max upper boundary on 1
        upper_values.append(min(ref_group_value * fairness_threshold, 1))

        lower_values.append(ref_group_value / fairness_threshold)

    # Convert to uppercase to match bubbles' Y axis
    metrics_labels = [metric.upper() for metric in metrics]

    threshold_df = pd.DataFrame(
        {
            "lower_end": 0,
            "upper_end": 1,
            "lower_threshold_value": lower_values,
            "upper_threshold_value": upper_values,
            "metric": metrics_labels,
        }
    )

    # RULES
    upper_threshold_rules = __draw_threshold_rules(
        threshold_df, scales, "upper", accessibility_mode
    )
    lower_threshold_rules = __draw_threshold_rules(
        threshold_df, scales, "lower", accessibility_mode
    )

    # BANDS
    threshold_bands = __draw_threshold_bands(threshold_df, scales, accessibility_mode)

    # HELPER TEXT
    if concat_chart:
        return lower_threshold_rules + upper_threshold_rules + threshold_bands
    else:
        threshold_text = __draw_threshold_text(
            fairness_threshold, ref_group, chart_height, accessibility_mode
        )

        return (
            lower_threshold_rules
            + upper_threshold_rules
            + threshold_bands
            + threshold_text
        )


def __draw_bubbles(
    plot_table,
    metrics,
    ref_group,
    scales,
    selection,
):
    """Draws the bubbles for all metrics."""

    # X AXIS GRIDLINES
    axis_values = [0.25, 0.5, 0.75]
    x_axis = alt.Axis(
        values=axis_values, ticks=False, domain=False, labels=False, title=None
    )

    # COLOR
    bubble_color_encoding = alt.condition(
        selection,
        alt.Color("attribute_value:N", scale=scales["color"], legend=None),
        alt.value(Bubble.color_faded),
    )

    # CHART INITIALIZATION
    bubble_centers = alt.Chart().mark_point()
    bubble_areas = alt.Chart().mark_circle()

    plot_table["tooltip_group_size"] = plot_table.apply(
        lambda row: get_tooltip_text_group_size(
            row["group_size"], row["total_entities"]
        ),
        axis=1,
    )
    # LAYERING THE METRICS
    for metric in metrics:
        # TOOLTIP
        plot_table[f"tooltip_disparity_explanation_{metric}"] = plot_table.apply(
            lambda row: get_tooltip_text_disparity_explanation(
                row[f"{metric}_disparity_scaled"],
                row["attribute_value"],
                metric,
                ref_group,
            ),
            axis=1,
        )

        bubble_tooltip_encoding = [
            alt.Tooltip(field="attribute_value", type="nominal", title="Group"),
            alt.Tooltip(field="tooltip_group_size", type="nominal", title="Group Size"),
            alt.Tooltip(
                field=f"tooltip_disparity_explanation_{metric}",
                type="nominal",
                title="Disparity",
            ),
            alt.Tooltip(
                field=f"{metric}",
                type="quantitative",
                format=".2f",
                title=f"{metric}".upper(),
            ),
        ]

        # BUBBLE CENTERS
        trigger_centers = alt.selection_multi(empty="all", fields=["attribute_value"])

        bubble_centers += (
            alt.Chart(plot_table)
            .transform_calculate(metric_variable=f"'{metric.upper()}'")
            .mark_point(filled=True, size=Bubble.center_size)
            .encode(
                x=alt.X(f"{metric}:Q", scale=scales["x"], axis=x_axis),
                y=alt.Y("metric_variable:N", scale=scales["y"], axis=no_axis()),
                tooltip=bubble_tooltip_encoding,
                color=bubble_color_encoding,
                shape=alt.Shape(
                    "attribute_value:N", scale=scales["shape"], legend=None
                ),
            )
            .add_selection(trigger_centers)
        )

        # BUBBLE AREAS
        trigger_areas = alt.selection_multi(empty="all", fields=["attribute_value"])

        bubble_areas += (
            alt.Chart(plot_table)
            .mark_circle(opacity=Bubble.opacity)
            .transform_calculate(metric_variable=f"'{metric.upper()}'")
            .encode(
                x=alt.X(f"{metric}:Q", scale=scales["x"], axis=x_axis),
                y=alt.Y("metric_variable:N", scale=scales["y"], axis=no_axis()),
                tooltip=bubble_tooltip_encoding,
                color=bubble_color_encoding,
                size=alt.Size("group_size:Q", legend=None, scale=scales["bubble_size"]),
            )
            .add_selection(trigger_areas)
        )

    return bubble_areas + bubble_centers


def get_metric_bubble_chart_components(
    plot_table,
    metrics,
    ref_group,
    global_scales,
    selection,
    fairness_threshold,
    chart_height,
    chart_width,
    accessibility_mode=False,
    concat_chart=False,
):
    """Creates the components necessary to plot the metric chart."""

    # COMPUTE SCALED DISPARITIES
    for metric in metrics:
        plot_table[f"{metric}_disparity_scaled"] = plot_table.apply(
            lambda row: transform_ratio(row[f"{metric}_disparity"]), axis=1
        )

    # POSITION SCALES
    position_scales = __get_position_scales(
        metrics, chart_height, chart_width, concat_chart
    )
    scales = dict(global_scales, **position_scales)

    # RULES
    horizontal_rules = __draw_metrics_rules(metrics, scales, concat_chart)
    vertical_domain_rules = __draw_domain_rules(scales)

    # LABELS
    x_ticks_labels = __draw_x_ticks_labels(scales, chart_height)

    # BUBBLES - CENTERS & AREAS
    bubbles = __draw_bubbles(
        plot_table,
        metrics,
        ref_group,
        scales,
        selection,
    )

    # THRESHOLD, BANDS & ANNOTATION
    if fairness_threshold is not None:
        threshold_elements = __get_threshold_elements(
            plot_table,
            metrics,
            fairness_threshold,
            ref_group,
            scales,
            chart_height,
            concat_chart,
            accessibility_mode,
        )
        # ASSEMBLE CHART WITH THRESHOLD
        main_chart = (
            horizontal_rules
            + vertical_domain_rules
            + x_ticks_labels
            + threshold_elements
            + bubbles
        )
    # ASSEMBLE CHART WITHOUT THRESHOLD
    else:
        main_chart = horizontal_rules + vertical_domain_rules + x_ticks_labels + bubbles

    return main_chart


def plot_metric_bubble_chart(
    disparity_df,
    metrics_list,
    attribute,
    fairness_threshold=1.25,
    chart_height=None,
    chart_width=Metric_Chart.full_width,
    accessibility_mode=False,
):
    """Draws bubble chart to visualize the values of the selected metrics for a given attribute.

    :param disparity_df: a dataframe generated by the Aequitas Bias class
    :type disparity_df: pandas.core.frame.DataFrame
    :param metrics_list: a list of the metrics of interest
    :type metrics_list: list
    :param attribute: an attribute to plot
    :type attribute: str
    :param fairness_threshold: a value for the maximum allowed disparity, defaults to 1.25
    :type fairness_threshold: float, optional
    :param chart_height: a value (in pixels) for the height of the chart
    :type chart_height: int, optional
    :param chart_width: a value (in pixels) for the width of the chart
    :type chart_width: int, optional
    :param accessibility_mode: a switch for the display of more accessible visual elements, defaults to False
    :type accessibility_mode: bool, optional

    :return: the full metrics chart
    :rtype: Altair chart object
    """
    (
        plot_table,
        metrics,
        ref_group,
        global_scales,
        chart_height,
        chart_width,
        selection,
    ) = Initializer.prepare_bubble_chart(
        disparity_df,
        metrics_list,
        attribute,
        fairness_threshold,
        chart_height,
        chart_width,
        Metric_Chart,
        accessibility_mode,
    )
    # GET MAIN CHART COMPONENTS
    main_chart = get_metric_bubble_chart_components(
        plot_table,
        metrics,
        ref_group,
        global_scales,
        selection,
        fairness_threshold,
        chart_height,
        chart_width,
        accessibility_mode,
    )

    # ADD LEGEND
    legend = draw_legend(global_scales, selection, chart_width)
    full_chart = main_chart + legend

    # FINALIZE CHART
    metric_chart = (
        full_chart.configure_view(strokeWidth=0)
        .configure_axisLeft(
            labelFontSize=Metric_Axis.label_font_size,
            labelColor=Metric_Axis.label_color,
            labelFont=FONT,
        )
        .properties(
            height=chart_height,
            width=chart_width,
            title=f"Absolute values by {attribute.title()}",
            padding=Metric_Chart.full_chart_padding,
        )
        .configure_title(
            align="center",
            baseline="middle",
            font=FONT,
            fontWeight=Chart_Title.font_weight,
            fontSize=Chart_Title.font_size,
            color=Chart_Title.font_color,
        )
        .resolve_scale(y="independent", size="independent")
    )

    return metric_chart



class Chart:
    full_chart_padding = {"top": 15, "bottom": 10}
    padding_x = 0.05
    padding_y = 0.1


class Summary_Chart(Chart):
    vertical_header_height = 45
    line_height = 85
    min_line_height = 50
    max_line_height = 150
    line_height_ratio = 0.65
    horizontal_header_width = 80
    column_width = 240
    min_column_width = 150
    max_column_width = 300
    column_width_ratio = 0.75


class Disparity_Chart(Chart):
    vertical_header_height = 100
    line_height = 150
    min_line_height = 80
    full_width = 800
    min_width = 380


class Metric_Chart(Chart):
    vertical_header_height = 50
    line_height = 125
    min_line_height = 80
    full_width = 800
    min_width = 380


class Concat_Chart(Chart):
    vertical_header_height = 110
    line_height = 75
    min_line_height = 50
    full_width = 760
    spacing = 0
    min_width = Metric_Chart.min_width + Disparity_Chart.min_width + spacing


class XY_Chart(Chart):
    full_side = 450
    min_side = 200
    padding_y = 0.05



import logging
import warnings
from aequitas.plotting import assemble_ref_groups
import numpy as np
import pandas as pd
from scipy import stats

logging.getLogger(__name__)

__author__ = "Rayid Ghani, Pedro Saleiro <saleiro@uchicago.edu>, Loren Hinkson"
__copyright__ = "Copyright \xa9 2018. The University of Chicago. All Rights Reserved."

class Bias(object):
    """
    """
    default_key_columns = ('model_id', 'score_threshold', 'attribute_name')
    all_group_metrics = ('ppr', 'pprev', 'precision', 'fdr', 'for', 'fpr',
                         'fnr', 'tpr', 'tnr', 'npv')
    non_attr_cols = ('score', 'model_id', 'as_of_date', 'entity_id', 'rank_abs',
                     'rank_pct', 'id', 'label_value')

    def __init__(self, key_columns=default_key_columns, sample_df=None,
                 non_attr_cols=non_attr_cols,
                 input_group_metrics=all_group_metrics, fill_divbyzero=None):
        """

        :param key_columns: optional, key identifying columns for grouping
            variables and bias metrics in intermediate joins. Defaults are
            'model_id', 'score_threshold', 'attribute_name'.
        :param input_group_metrics: List of absolute bias metrics to calculate
        :param fill_divbyzero: optional, fill value to use when divided by
            zero. Default is None.
        """
        self.key_columns = list(key_columns)
        self.input_group_metrics = list(input_group_metrics)

        if not fill_divbyzero:
            self.fill_divbyzero = 10.00000
        else:
            self.fill_divbyzero = fill_divbyzero
        self.non_attr_cols = non_attr_cols
        self.significance_cols = set(self.input_group_metrics).union({'label_value', 'score'})

    def get_disparity_min_metric(self, df, original_df, key_columns=None,
                                 input_group_metrics=None, fill_divbyzero=None,
                                 check_significance=False,  alpha = 5e-2,
                                 mask_significance = True, label_score_ref='fpr',
                                 selected_significance=False):
        """
        Calculates disparities between groups for the predefined list of
        group metrics using the group with the minimum value for each absolute
        bias metric as the reference group (denominator).

        :param df: output dataframe of Group class get_crosstabs() method.
        :param original_df: a dataframe of sample features and model results.
            Includes a required 'score 'column and possible 'label_value' column.
        :param key_columns: optional, key identifying columns for grouping
            variables and bias metrics in intermediate joins. Defaults are
            'model_id', 'score_threshold', 'attribute_name'.
        :param input_group_metrics: optional, columns list corresponding to
            the group metrics for which we want to calculate disparity values
        :param fill_divbyzero: optional, fill value to use when divided by
            zero. Default is None.
        :param check_significance: whether to determine statistical signifance
            for disparity metrics. Default is False.
        :param selected_significance: specific measures (beyond label_value and
            score) to which to limit statistical significance calculations when
            check_significance is True. Default is False, i.e., significance
            will be calculated for all metrics.
        :param alpha: statistical significance level to use in significance
            determination. Default is 5e-2 (0.05).
        :param mask_significance: whether to display a T/F mask over calculated
            p-values from statistical significance determination. Default is True.
        :param label_score_ref: default reference group to use for score and
            label_value statistical significance calculations.
        :return: A dataframe with same number of rows as the input (crosstab)
            with additional disparity metrics columns and ref_group_values
            for each metric.
        """

        print('get_disparity_min_metric()')
        # record df column order
        original_cols = df.columns

        if not key_columns:
            key_columns = original_cols.intersection(self.key_columns).tolist()
        if not input_group_metrics:
            input_group_metrics = self.input_group_metrics
        if not fill_divbyzero:
            fill_divbyzero = self.fill_divbyzero

        for group_metric in input_group_metrics:

            try:
                # this groupby is being called every cycle. maybe we can create
                # a list of df_groups and merge df at the end? it can not be
                # simply put outside the loop(the merge...)
                idxmin = df.groupby(key_columns)[group_metric].idxmin()

                # if entire column for a group metric is NaN, cast min value index
                # column to the same index as that of any other group for that attribute
                if any(np.isnan(val) for val in idxmin.values):
                    if len(idxmin) >= 1:
                        idxmin_not_nan = df.reset_index().groupby(key_columns)['index'].min()
                        idxmin.loc[idxmin.isna()] = pd.merge(
                            left=idxmin.loc[idxmin.isna()], right=idxmin_not_nan,
                            left_index=True, right_index=True, how='inner',
                        )['index'].values
                    else:
                        raise Exception(f"A minimum value for group_metric "
                                      f"{group_metric} could not be calculated.")

                df_min_idx = df.loc[idxmin]

                # but we also want to get the group_value of the reference group
                # for each bias metric
                df_to_merge = pd.DataFrame()
                df_to_merge[key_columns + [group_metric + '_disparity', group_metric +
                                           '_ref_group_value']] = \
                    df_min_idx[key_columns + [group_metric, 'attribute_value']]

            except KeyError:
                raise KeyError(
                    'get_bias_min_metric:: one of the following columns is not '
                    'on the input dataframe : model_id, parameter, attribute_name '
                    'or any of the input_group_metrics '
                    'list')

            df = df.merge(df_to_merge, on=key_columns)
            # creating disparity by dividing each group metric value by the
            # corresponding min value from the groups of the target attribute
            df[group_metric + '_disparity'] = \
                df[group_metric] / df[group_metric + '_disparity']
            # We are capping the disparity values to 10.0 when divided by zero...
        df = df.replace(np.inf, fill_divbyzero)

        if not check_significance:
            return df

        else:
            # add statistical_significance
            if not selected_significance:
                selected_significance = self.input_group_metrics

                # only proceed with columns actually in dataframe
            selected_significance = set(original_cols.intersection(selected_significance))

            # always includes label and score significance
            selected_significance = selected_significance.union({'label_value', 'score'})

            ref_groups_dict = assemble_ref_groups(df, ref_group_flag='_ref_group_value',
                                                  specific_measures=selected_significance,
                                                  label_score_ref=label_score_ref)

            attr_cols = df['attribute_name'].unique()
            # run significance method on bias-augmented crosstab based on false
            # positives, false negatives, scores, and label values in original df
            self._get_statistical_significance(
                original_df, df, ref_dict=ref_groups_dict, score_thresholds=None,
                attr_cols=attr_cols, alpha=5e-2, selected_significance=selected_significance)

            # if specified, apply T/F mask to significance columns
            if mask_significance:
                significance_cols = df.columns[df.columns.str.contains('_significance')]
                truemask = df.loc[:, significance_cols] < alpha
                falsemask = df.loc[:, significance_cols] >= alpha

                df.loc[:, significance_cols] = np.select(
                    [truemask, falsemask], [True, False], default=None)

            # order new calculated metric columns: disparity, ref_group, then
            # significance for each
            base_sig=['label_value_significance', 'score_significance']

            new_cols = sorted(
                list(set(df.columns) - set(original_cols) - set(base_sig))
            )

            return df[original_cols.tolist() + base_sig + new_cols]


    def get_disparity_major_group(self, df, original_df, key_columns=None,
                                  input_group_metrics=None,
                                  fill_divbyzero=None, check_significance=False,
                                  alpha = 5e-2, mask_significance=True,
                                  selected_significance=False):
        """
        Calculates disparities between groups for the predefined list of group
        metrics using the majority group within each attribute as the reference
        group (denominator).

        :param df: output dataframe of Group class get_crosstabs() method.
        :param original_df: a dataframe of sample features and model results.
            Includes a required 'score 'column and possible 'label_value' column.
        :param key_columns: optional, key identifying columns for grouping
            variables and bias metrics in intermediate joins. Defaults are
            'model_id', 'score_threshold', 'attribute_name'.
        :param input_group_metrics: optional, columns list corresponding to
            the group metrics for which we want to calculate disparity values
        :param fill_divbyzero: optional, fill value to use when divided by
            zero. Default is None.
        :param check_significance: whether to determine statistical signifance
            for disparity metrics. Default is False.
        :param selected_significance: specific measures (beyond label_value and
            score) to which to limit statistical significance calculations when
            check_significance is True. Default is False, i.e., significance
            will be calculated for all metrics.
        :param alpha: statistical significance level to use in significance
            determination. Default is 5e-2 (0.05).
        :param mask_significance: whether to display a T/F mask over calculated
            p-values from statistical significance determination. Default is True.
        :return: A dataframe with same number of rows as the input (crosstab)
            with additional disparity metrics columns and ref_group_values
            for each metric.
        """
        print('get_disparity_major_group()')
        # record df column order
        original_cols = df.columns

        if not key_columns:
            key_columns = original_cols.intersection(self.key_columns).tolist()
        if not input_group_metrics:
            input_group_metrics = self.input_group_metrics
        if not fill_divbyzero:
            fill_divbyzero = self.fill_divbyzero

        try:
            df_major_group = df.loc[df.groupby(key_columns)['group_size'].idxmax()]
        except KeyError:
            raise KeyError('get_bias_major_group:: one of the following columns '
                          'is not on the input dataframe : model_id, parameter, '
                          'attribute_name, group_size')

        disparity_metrics = [col + '_disparity' for col in input_group_metrics]
        df_to_merge = pd.DataFrame()

        # we created the df_to_merge has a subset of the df_ref_group containing
        # the target ref group values which are now labeled as _disparity but
        # we still need to perform the division
        df_to_merge[key_columns + disparity_metrics] = df_major_group[
            key_columns + input_group_metrics]

        # we now need to create the ref_group_value columns in the df_to_merge
        for col in input_group_metrics:
            df_to_merge[col + '_ref_group_value'] = df_major_group['attribute_value']
        df = df.merge(df_to_merge, on=key_columns)
        df[disparity_metrics] = df[input_group_metrics].divide(df[disparity_metrics].values)

        # We are capping the disparity values to 10.0 when divided by zero...
        df = df.replace(np.inf, fill_divbyzero)

        # when there is a zero in the numerator and a zero in denominator it is
        # considered NaN after division, so if 0/0 we assume 1.0 disparity
        # (they are the same...)

        # default is to use the same ref groups as df, need to add functionality to
        # compile ref_groups_dict based on a passed ref group for a given measure
        if not check_significance:
            return df

        else:
            if not selected_significance:
                selected_significance = self.input_group_metrics

            # only proceed with columns actually in dataframe
            selected_significance = set(original_cols.intersection(selected_significance))

            # always includes label and score significance
            selected_significance = selected_significance.union({'label_value', 'score'})

            ref_groups_dict = assemble_ref_groups(df, ref_group_flag='_ref_group_value',
                                                  specific_measures=selected_significance,
                                                  label_score_ref=None)

            ref_groups_dict = assemble_ref_groups(df, ref_group_flag='_ref_group_value',
                                                  specific_measures=selected_significance,
                                                  label_score_ref=None)

            attr_cols = df['attribute_name'].unique()

            for attribute in attr_cols:
                largest_group = df_major_group.loc[df_major_group['attribute_name'] == attribute,
                                                   'attribute_value'].values.tolist()[0]
                ref_groups_dict[attribute]['label_value'] = largest_group
                ref_groups_dict[attribute]['score'] = largest_group


            # run significance method on bias-augmented crosstab based on false
            # positives, false negatives, scores, and label values in original df
            self._get_statistical_significance(
                original_df, df, ref_dict=ref_groups_dict, score_thresholds=None,
                attr_cols=attr_cols, alpha=5e-2, selected_significance=selected_significance)

            # if specified, apply T/F mask to significance columns
            if mask_significance:
                significance_cols = df.columns[df.columns.str.contains('_significance')]
                truemask = df.loc[:, significance_cols] < alpha
                falsemask = df.loc[:, significance_cols] >= alpha

                df.loc[:, significance_cols] = np.select(
                    [truemask, falsemask], [True, False], default=None)

            # check what new disparity columns are and order as disparity,
            # ref_group, significance for each
            base_sig=['label_value_significance', 'score_significance']

            new_cols = sorted(
                list(set(df.columns) - set(original_cols) - set(base_sig))
            )

            return df[original_cols.tolist() + base_sig + new_cols]


    def _verify_ref_groups_dict_len(self, df, ref_groups_dict):
        if len(ref_groups_dict) != len(df['attribute_name'].unique()):
            raise ValueError

    def _verify_ref_group_loc(self, group_slice):
        if len(group_slice) < 1:
            raise ValueError

    def get_disparity_predefined_groups(self, df, original_df, ref_groups_dict,
                                        key_columns=None,
                                        input_group_metrics=None,
                                        fill_divbyzero=None,
                                        check_significance=False, alpha=5e-2,
                                        mask_significance=True,
                                        selected_significance=False):
        """
        Calculates disparities between groups for the predefined list of group
        metrics using a predefined reference group (denominator) value for each
        attribute.

        :param df: output dataframe of Group class get_crosstabs() method.
        :param original_df: dataframe of sample features and model results.
            Includes a required 'score 'column and possible 'label_value' column.
        :param ref_groups_dict: dictionary of format: {'attribute_name': 'attribute_value', ...}
        :param key_columns: optional, key identifying columns for grouping
            variables and bias metrics in intermediate joins. Defaults are
            'model_id', 'score_threshold', 'attribute_name'.
        :param input_group_metrics: optional, columns list corresponding to
            the group metrics for which we want to calculate disparity values
        :param fill_divbyzero: optional, fill value to use when divided by
            zero. Default is None.
        :param check_significance: whether to determine statistical signifance
            for disparity metrics. Default is False.
        :param selected_significance: specific measures (beyond label_value and
            score) to which to limit statistical significance calculations when
            check_significance is True. Default is False, i.e., significance
            will be calculated for all metrics.
        :param alpha: statistical significance level to use in significance
            determination. Default is 5e-2 (0.05).
        :param mask_significance: whether to display a T/F mask over calculated
            p-values from statistical significance determination. Default is True.
        :return: A dataframe with same number of rows as the input (crosstab)
            with additional disparity metrics columns and ref_group_values
            for each metric.
        """
        print('get_disparity_predefined_group()')
        # record df column order
        original_cols = df.columns

        if not key_columns:
            key_columns = original_cols.intersection(self.key_columns).tolist()
        if not input_group_metrics:
            input_group_metrics = self.input_group_metrics
        if not fill_divbyzero:
            fill_divbyzero = self.fill_divbyzero

        try:
            self._verify_ref_groups_dict_len(df, ref_groups_dict)
        except ValueError:
            raise ValueError('Bias.get_disparity_predefined_groups(): the number of '
                          'predefined group values to use as reference is less '
                          'than the actual number of attributes in the input '
                          'dataframe.')

        df_ref_group = pd.DataFrame()
        try:
            for key, val in ref_groups_dict.items():
                group_slice = df.loc[(df['attribute_name'] == key) &
                                     (df['attribute_value'] == val)]
                self._verify_ref_group_loc(group_slice)
                df_ref_group = pd.concat([df_ref_group, group_slice])
        except KeyError:
            raise KeyError('get_disparity_predefined_groups(): reference groups '
                          'and values provided do not exist as columns/values '
                          'in the input dataframe.(Note: check for syntax errors)')
        except ValueError:
            raise ValueError('get_disparity_predefined_groups(): reference groups '
                           'and values provided do not exist as columns/values '
                           'in the input dataframe.(Note: check for syntax errors)')

        disparity_metrics = [col + '_disparity' for col in input_group_metrics]
        df_to_merge = pd.DataFrame()

        # we created the df_to_merge has a subset of the df_ref_group containing
        # the target ref group values which are now labeled as _disparity but
        # we still need to perform the division
        df_to_merge[key_columns + disparity_metrics] = df_ref_group[
            key_columns + input_group_metrics]

        # we now need to create the ref_group_value columns in the df_to_merge
        for col in input_group_metrics:
            df_to_merge[col + '_ref_group_value'] = df_ref_group['attribute_value']
        df = df.merge(df_to_merge, on=key_columns)
        df[disparity_metrics] = df[input_group_metrics].divide(df[disparity_metrics].values)

        # We are capping the disparity values to 10.0 when divided by zero...
        df = df.replace(np.inf, fill_divbyzero)
        if not check_significance:
            return df

        else:
            if not selected_significance:
                selected_significance = self.input_group_metrics

            # only proceed with columns actually in dataframe
            selected_significance = set( original_cols.intersection(selected_significance) )

            # always includes label and score significance
            selected_significance = selected_significance.union({'label_value', 'score'})

            # compile dictionary of reference groups based on bias-augmented crosstab
            full_ref_dict = {}

            # for predefined groups, use the largest of the predefined groups as
            # ref group for score and label value
            # key is an attribute_name, val is an attribute_value
            for key, val in ref_groups_dict.items():
                full_ref_dict[key] = {'label_value': val,
                                      'score': val}
                for measure in selected_significance:
                    full_ref_dict[key][measure] = val

            # run significance method on bias-augmented crosstab based on false
            # positives, false negatives, scores, and label values in original df
            self._get_statistical_significance(
                original_df, df, ref_dict=full_ref_dict, score_thresholds=None,
                attr_cols=None, alpha=5e-2, selected_significance=selected_significance)

            # if specified, apply T/F mask to significance columns
            if mask_significance:
                significance_cols = df.columns[df.columns.str.contains('_significance')]
                truemask = df.loc[:, significance_cols] < alpha
                falsemask = df.loc[:, significance_cols] >= alpha

                df.loc[:, significance_cols] = np.select(
                    [truemask, falsemask], [True, False], default=None)

            # check what new disparity columns are and order as disparity,
            # ref_group, significance for each
            base_sig=['label_value_significance', 'score_significance']

            new_cols = sorted(
                list(set(df.columns) - set(original_cols) - set(base_sig))
            )

            return df[original_cols.tolist() + base_sig + new_cols]



    @staticmethod
    def _get_measure_sample(original_df, attribute, measure):
        """
        Helper function for _get_statistical_significance() (via
        _calculate_significance() function). Convert dataframe to samples for
        given attribute group.

        :param original_df: a dataframe containing a required raw 'score' column
            and possible raw 'label_value' column.
        :param attribute: Attribute of interest in dataset (ex: race, sex, age
            category)
        :param measure: Metric of interest for which to calculate significance
            (false positives, false negatives, score, label_value)

        :return: A dictionary of binary 'samples' for each attribute group
        """
        return original_df.groupby(attribute).apply(
            lambda f: f.loc[f[measure].notnull(), measure].values.tolist()).to_dict()


    @staticmethod
    def _check_equal_variance(sample_dict, ref_group, alpha=5e-2):
        """
        Helper function for _get_statistical_significance() (via
        _calculate_significance() function).

        :param sample_dict: dictionary of binary samples for equal variance
            comparison.
        :param ref_group: Group to use as reference group for statistical
            significance calculation.
        :param alpha: Level at which to determine statistical significance.
            Default is 5e-2 (0.05).

        :return: Dictionary indicating whether each group has equal variance
        (in comparison with reference group)
        """

        # Immediately set ref_group status for equal variance (with itself): True

        eq_variance = {ref_group: True}

        for attr_value, sample in sample_dict.items():
            # make default normality_p value (only used when len(sample) < 20)
            # large enough that it is always greater than alpha
            normality_p = np.inf

            # skew test requires at least 20 samples
            if len(sample) > 20:
                _, normality_p = stats.normaltest(sample, axis=None, nan_policy='omit')

            # if tested normality is False or less than 8 samples, use levene
            # test to check equal variance between groups
            if normality_p < alpha or len(sample) <= 20:
                # if ref_group is not normal, can't use f-test or bartlett test
                # for any samples, so check for equal variance against ref_group
                # using levene test for all groups and return dict
                # (since includes all groups)
                if attr_value == ref_group:
                    for group, sample_list in sample_dict.items():
                        _, equal_variance_p = stats.levene(sample_dict[ref_group],
                                                           sample_list,
                                                           center='median')

                        eq_variance[group] = equal_variance_p >= alpha


                    return eq_variance

                # if a non-ref group is not normal, can't use f-test or bartlett
                # for that group, check for equal variance (against ref_group)
                # using levene test and add result to dictionary
                _, equal_variance_p = stats.levene(
                    sample_dict[ref_group], sample, center='median')

                eq_variance[attr_value] = equal_variance_p >= alpha

        # for all normally distributed non-ref groups, use bartlett test to
        # check for equal variance (against ref_group). Add results to dict
        untested_groups = sample_dict.keys() - eq_variance.keys() - set(ref_group)
        untested = {key: val for (key, val) in sample_dict.items()
                    if key in untested_groups}
        for attr_value, sample_list in untested.items():
            _, equal_variance_p = stats.bartlett(sample_dict[ref_group], sample_list)

            eq_variance[attr_value] = equal_variance_p >= alpha

        return eq_variance


    @classmethod
    def _calculate_significance(cls, original_df, disparity_df, attribute,
                               measure, ref_dict, alpha=5e-2):
        """
        Helper function for _get_statistical_significance. Pulls samples from
        original df, checks for equal variance between population groups and
        reference group, runs t-test between groups and reference group, and
        adds p-values to disparity_df for a given measure (ex: false positives)

        :param original_df: a dataframe containing a required raw 'score' column
            and possible raw 'label_value' column.
        :param disparity_df: Expansion of get_crosstabs() output with additional
            columns for disparity metrics and each metric's reference group to
            which significance must be added
        :param attribute: Attribute for which to calculate statistical
            significance of a measure
        :param measure: Measure for which to calculate statistical significance
        :param ref_dict: Dictionary indicating reference group for each
            attribute/ measure combination
        :param alpha: Level at which to determine statistical significance.
            Default is 5e-2 (0.05).

        :return: dataframe with same number of rows as the input (crosstab)
            but with additional disparity metrics columns, ref_group_values, and
            statistical significance (p-values) of specific metrics.

        """
        binaries_lookup = {'label_value': 'label_value', 'binary_fpr': 'fpr',
                           'binary_tpr': 'tpr', 'binary_tnr': 'tnr',
                           'binary_fnr': 'fnr', 'binary_score': 'score',
                           'binary_precision': 'precision', 'binary_npv': 'npv',
                           'binary_for': 'for', 'binary_fdr': 'fdr',
                           'binary_ppr': 'ppr', 'binary_pprev': 'pprev'
                            }

        ref_group = ref_dict[attribute][binaries_lookup.get(measure)]

        # create dictionary of "samples" (binary values for false positive,
        # false negative, label value, score) based on original data frame
        sample_dict = cls._get_measure_sample(original_df=original_df,
                                              attribute=attribute, measure=measure)


        # run SciPy equal variance tests between each group and a given
        # reference group, store results in dictionary to pass to statistical
        # significance tests
        eq_variance_dict = cls._check_equal_variance(sample_dict=sample_dict,
                                                    ref_group=ref_group,
                                                    alpha=alpha)
        # run SciPy statistical significance test between each group and
        # reference group
        for attr_val, eq_var in sample_dict.items():

            _, difference_significance_p = stats.ttest_ind(
                sample_dict[ref_group],
                sample_dict[attr_val],
                axis=None,
                equal_var=eq_variance_dict[attr_val],
                nan_policy='omit')

            measure = measure.replace('binary_', '')

            # add column to crosstab to indicate statistical significance
            disparity_df.loc[disparity_df['attribute_value'] == attr_val,
                             measure + '_significance'] = difference_significance_p

        return disparity_df



    @classmethod
    def _get_statistical_significance(cls, original_df, disparity_df, ref_dict,
                                     score_thresholds=None,
                                     attr_cols=None, alpha=5e-2,
                                     selected_significance=False):
        """

        :param original_df: a dataframe containing a required raw 'score' column
            and possible raw 'label_value' column.
        :param disparity_df: Expansion of get_crosstabs() output with additional
            columns for disparity metrics and each metric's reference group to
            which significance must be added
        :param ref_dict: Dictionary indicating reference group for each
            attribute/ measure combination
        :param score_thresholds: a dictionary { 'rank_abs':[] , 'rank_pct':[], 'score':[] }
        :param model_id: (Future functionality) ID(s) of models for which to check
            statistical significance
        :param attr_cols: Columns indicating attribute values in original_df
        :param alpha: Level at which to determine statistical significance.
            Default is 5e-2 (0.05).

        :return: dataframe with same number of rows as the input (crosstab)
            but with additional disparity metrics columns, ref_group_values, and
            statistical significance (p-values) of specific metrics.
        """
        if 'label_value' not in original_df.columns:
            raise ValueError(
                "Column 'label_value' not in dataframe. Label values are "
                "required for computing statistical significance of supervised "
                "metrics.")

        if attr_cols is None:
            non_attr_cols = [
                'id', 'model_id', 'entity_id', 'score', 'label_value',
                'rank_abs', 'rank_pct']

            # index of the columns that are attributes
            attr_cols = original_df.columns[~original_df.columns.isin(non_attr_cols)]

        # check if all attr_cols exist in df
        if set(attr_cols) - set(original_df.columns):
            raise ValueError(
                f"Not all attribute columns provided '{attr_cols}' exist in "
                f"input dataframe.")

        # check if all columns are strings:
        non_string_cols = \
            original_df.columns[
                (original_df.dtypes != object) &
                (original_df.dtypes != str) &
                (original_df.columns.isin(attr_cols))]

        binary_inclusions = {f'binary_{col}' for col in (selected_significance or
                                                         cls.significance_cols)
                             if col != 'label_value'}

        # check whether necessary binary columns already created
        if not non_string_cols.empty:
            # check whether error message is required based on if only binary columns
            if not all('binary_' in col for col in non_string_cols):
                raise ValueError(
                    'get_statistical_significance: statistical significance was '
                    'not calculated. There are non-string cols within attr_cols.')


            # skip binary column creation if all necessary columns already created
            elif not binary_inclusions - set(non_string_cols):
                measures = set(original_df.columns[original_df.columns.str.contains('binary_')])
                measures = measures.union({'label_value'})

                for attribute in set(attr_cols) - measures:
                    for measure in measures:

                        # only calculate significance if in selected_significance
                        if (measure in selected_significance) or (
                                measure.replace('binary_', '') in selected_significance):
                            cls._calculate_significance(original_df, disparity_df,
                                                        attribute, measure, ref_dict=ref_dict,
                                                        alpha=alpha)

                return disparity_df

            else:
                # drop old binary columns and generate those in binary_inclusions
                # as though non_string_cols was empty
                drop_cols = list(original_df.columns[original_df.columns.str.contains('binary_')])
                drop_cols += list(original_df.columns[original_df.columns.str.contains('rank_')])
                original_df = original_df.drop(drop_cols, axis=1)

        # create binary columns if non_string_cols is empty or binaries dropped

        # if no score_thresholds are provided, we assume that rank_abs equals
        # the number  of 1s in the score column; it also serves as flag to set
        # parameter to 'binary'
        count_ones = None
        if not score_thresholds:
            original_df.loc[:, 'score'] = original_df.loc[:,'score'].astype(float)

            count_ones = original_df['score'].value_counts().get(1.0, 0)
            score_thresholds = {'rank_abs': [count_ones]}

        original_df.sort_values('score', ascending=False, inplace=True)
        original_df.loc[:,'rank_abs'] = range(1, len(original_df) + 1)
        original_df.loc[:,'rank_pct'] = original_df.loc[:,'rank_abs'] / len(original_df)

        # Define formula for binary false positive, false negative, binary score
        binary_false_pos = lambda rank_col, label_col, thres: lambda x: (
            (x[rank_col] <= thres) & (x[label_col] == 0)).astype(int)

        binary_false_neg = lambda rank_col, label_col, thres: lambda x: (
            (x[rank_col] > thres) & (x[label_col] == 1)).astype(int)

        binary_score = lambda rank_col, label_col, thres: lambda x: (
                x[rank_col] <= thres).astype(int)

        binary_col_functions = {'binary_score': binary_score,
                                'binary_fpr': binary_false_pos,
                                'binary_fnr': binary_false_neg
                                }

        for thres_unit, thres_values in score_thresholds.items():
            for thres_val in thres_values:
                for name, func in binary_col_functions.items():
                    func = func(thres_unit, 'label_value', thres_val)
                    original_df.loc[:, name] = original_df.apply(func, axis=1)

        # add columns for error-based significance
        # precision, tnr, fdr are based on false positives
        SIGNIF_BASES = {
            'binary_precision': 'binary_fpr',
            'binary_tnr': 'binary_fpr',
            'binary_fdr': 'binary_fpr',
            'binary_npv': 'binary_fnr',
            'binary_tpr': 'binary_fnr',
            'binary_for': 'binary_fnr',
            'binary_ppr': 'binary_score',
            'binary_pprev': 'binary_score'}

        # add columns for the rest of columns in dictionary keys
        # binary score, fnr, fpr already added above
        for col in (binary_inclusions - binary_col_functions.keys()):
            original_df.loc[:, col] = original_df.loc[:,SIGNIF_BASES[col]]

        # ensure only predicted positive values included in true positive
        # error based metrics, and only predicted negative values in false
        # positive error based metrics
        POSITIVE_ONLY = ['binary_fpr', 'binary_tnr', 'binary_precision', 'binary_fdr']
        NEGATIVE_ONLY = ['binary_fnr', 'binary_tpr', 'binary_npv', 'binary_for']

        FPR_BASED = [metric for metric in POSITIVE_ONLY if metric in binary_inclusions]
        FNR_BASED = [metric for metric in NEGATIVE_ONLY if metric in binary_inclusions]

        original_df.loc[original_df['binary_score'] == 0, FPR_BASED] = np.nan
        original_df.loc[original_df['binary_score'] == 1, FNR_BASED] = np.nan

        measures = set(original_df.columns[original_df.columns.str.contains('binary_')])
        measures = measures.union({'label_value'})

        for attribute in attr_cols:
            for measure in measures:

                # only calculate significance if in selected_significance
                if (measure in selected_significance) or (measure.replace('binary_', '') in selected_significance):

                    cls._calculate_significance(original_df, disparity_df,
                                                attribute, measure, ref_dict=ref_dict,
                                                alpha=alpha)
        return disparity_df


    def list_disparities(self, df):
        """
        View list of all calculated disparities in df
        """
        try:
            return list(df.columns[df.columns.str.contains('_disparity')])
        except KeyError:
            raise KeyError("No disparity columns found in dataframe. Tip: "
                            "make sure you have already passed the dataframe "
                            "through a 'get_disparity_' method.")


    def list_significance(self, df):
        """
        View list of all calculated disparities in df
        """
        try:
            return list(df.columns[df.columns.str.contains('_significance')])
        except KeyError:
            raise KeyError("No significance columns found in dataframe. Tip: "
                            "make sure you set the 'check_significance' parameter"
                            " to True in 'get_disparity_' method(s).")


    def list_absolute_metrics(self, df):
        """
        View list of all calculated absolute bias metrics in df
        """
        return list(set(self.input_group_metrics).intersection(df.columns))



from aequitas.plot.commons.style import sizes as Sizes
from aequitas.plot.commons.helpers import calculate_chart_size_from_elements, to_list

METRICS_LIST = [
    "tpr",
    "tnr",
    "for",
    "fdr",
    "fpr",
    "fnr",
    "npv",
    "ppr",
    "precision",
    "pprev",
    "prev",
    "group_size_pct",
]
NON_DISPARITY_METRICS_LIST = ["prev", "group_size_pct"]
DF_FIELDS = ["attribute_name", "attribute_value", "group_size", "total_entities"]

# DATA


def dataframe(df, metrics_list):
    """Validates if dataframe has all the necessary fieds to plot charts.

    :param user_input_df: a dataframe from Aequitas
    :type user_input_df: pandas.core.frame.DataFrame
    :param metrics_list: a list of valid metrics
    :type metrics_list: list
    """
    for field in DF_FIELDS:
        try:
            df[field]
        except KeyError:
            print(f'The dataframe does not contain the column "{field}"')
            raise

    for metric in metrics_list:
        try:
            df[metric]
        except KeyError:
            print(f'The dataframe does not contain the column "{metric}".')
            raise
        if metric not in NON_DISPARITY_METRICS_LIST:
            try:
                df[f"{metric}_disparity"]
            except KeyError:
                print(
                    f'The dataframe does not contain the column "{metric}_disparity".'
                )
                raise


def metrics(metrics):
    """Validates if  if there are no duplicated metrics, and if all metrics are valid fairness metrics.

    :param metrics: a metric or a list of metrics
    :type metrics: str or list 

    """

    metrics = to_list(metrics)

    if len(metrics) != len(set(metrics)):
        raise ValueError(
            f"There is at least one duplicated metric in the metrics list."
        )

    for metric in metrics:
        if metric not in METRICS_LIST:
            raise ValueError(f"{metric} is not a valid metric.")


def attributes(df, attributes):
    """Validates if attributes exists in the dataframe.

    :param df: a dataframe from Aequitas
    :type df: pandas.core.frame.DataFrame
    :param attributes: an attribute or a list of attributes
    :type attributes: str or list 
    """

    attributes = to_list(attributes)

    for attribute in attributes:
        attribute_df = df[df["attribute_name"] == attribute]

        if len(attribute_df) == 0:
            raise ValueError(
                f"Attribute name {attribute} does not exist in the dataframe."
            )


def fairness_threshold(fairness_threshold):
    """Checks if fairness_threshold is a number greater than 1.
    :param fairness_threshold: a threshold value
    :type fairness_threshold: number
    """
    if fairness_threshold is not None:
        try:
            if fairness_threshold <= 1:
                raise ValueError(f"Fairness Threshold value must be greater than 1.")
        except TypeError:
            print("Fairness Threshold value must be a number")
            raise


# SIZES


def chart_size_summary(size_constants, num_metrics, num_attributes):
    """Given sizing constants for the summary chart and the number of metrics and attributes, this function checks if there is enough area to render all elements and if there is a sensible aspect ratio.

    :param size_constants: size constants for Summary Chart size elements
    :type size_constants: dict
    :param num_metrics: number of metrics to be plotted (rows)
    :type num_metrics: int
    :param num_attributes: number of attributes to be plotted (columns)
    :type num_attributes: int
    """

    min_width = (
        Sizes.Summary_Chart.min_column_width
        * num_attributes
        / Sizes.Summary_Chart.column_width_ratio
    )
    max_width = (
        Sizes.Summary_Chart.max_column_width
        * num_attributes
        / Sizes.Summary_Chart.column_width_ratio
    )
    min_height = Sizes.Summary_Chart.min_line_height * num_metrics / 0.65
    max_height = Sizes.Summary_Chart.max_line_height * num_metrics / 0.65

    if size_constants["column_width"] < Sizes.Summary_Chart.min_column_width:
        raise ValueError(
            f"The chart width must be at least {min_width} to accommodate the {num_attributes} attributes to be plotted."
        )

    if size_constants["column_width"] > Sizes.Summary_Chart.max_column_width:
        raise ValueError(
            f"The chart width must be smaller than {max_width} to preserve a sensible aspect ratio."
        )
    if size_constants["line_height"] < Sizes.Summary_Chart.min_line_height:
        raise ValueError(
            f"The chart height must be at least {min_height} to accommodate the {num_metrics} metrics to be plotted."
        )
    if size_constants["line_height"] > Sizes.Summary_Chart.max_line_height:
        raise ValueError(
            f"The chart height must be smaller than {max_height} to preserve a sensible aspect ratio."
        )


def chart_size_bubble(width, height, chart_sizes, num_metrics):
    """Given parameters width and height for the bubble charts and the number of metrics, this function checks if there is enough area to render all elements and if there is a sensible aspect ratio.

    :param width: chart width
    :type width: int
    :param height: chart height
    :type height: int
    :param chart_sizes: size style defaults for the chart 
    :type chart_sizes: class
    :param num_metrics: number of metrics to be plotted
    :type num_metrics: int
    """
    min_height = calculate_chart_size_from_elements(
        chart_sizes.vertical_header_height, chart_sizes.min_line_height, num_metrics,
    )

    if width < chart_sizes.min_width:
        raise ValueError(
            f"The chart width must be at least {chart_sizes.min_width} to preserve a sensible aspect ratio."
        )

    if height < min_height:
        raise ValueError(
            f"The chart height must be at least {min_height} to accommodate the {num_metrics} metrics to be plotted."
        )


def chart_size_xy(width, height):
    """Given parameters width and height for the XY chart, this function checks if height is equal to width and if there is enough area to render all elements.

    :param width: chart width
    :type width: int
    :param height: chart height
    :type height: int
    """
    if width != height:
        raise ValueError(f"The chart width and height must be the same.")

    if width < Sizes.XY_Chart.min_side:
        raise ValueError(
            f"The chart width must be at least {Sizes.XY_Chart.min_side} to preserve a sensible aspect ratio."
        )




import math
import altair as alt
import pandas as pd

from aequitas.plot.commons.helpers import (
    no_axis,
    transform_ratio,
)
from aequitas.plot.commons.legend import draw_legend
from aequitas.plot.commons.scales import get_chart_size_range
from aequitas.plot.commons.tooltips import (
    get_tooltip_text_group_size,
    get_tooltip_text_disparity_explanation,
)
from aequitas.plot.commons.style.classes import (
    Metric_Axis,
    Axis,
    Annotation,
    Reference_Rule,
    Threshold_Rule,
    Threshold_Band,
    Bubble,
    Chart_Title,
)

from aequitas.plot.commons.style.text import FONT
from aequitas.plot.commons.style.sizes import Disparity_Chart
from aequitas.plot.commons import initializers as Initializer


# Altair 2.4.1 requires that all chart receive a dataframe, for charts that don't need it
# (like most annotations), we pass the following dummy dataframe to reduce the complexity of the resulting vega spec.
DUMMY_DF = pd.DataFrame({"a": [1, 1], "b": [0, 0]})


def __get_position_scales(
    plot_table, metrics, fairness_threshold, chart_height, chart_width
):
    """Computes the scales for x and y encodings to be used in the disparity bubble chart."""

    position_scales = dict()

    # DISPARITIES SCALE
    # RANGE
    x_range = get_chart_size_range(chart_width, Disparity_Chart.padding_x)

    # DOMAIN
    # Get max absolute disparity
    scaled_disparities_col_names = [f"{metric}_disparity_scaled" for metric in metrics]

    def max_column(x):
        return max(x.min(), x.max(), key=abs)

    max_disparities = plot_table[scaled_disparities_col_names].apply(max_column, axis=1)
    abs_max_disparity = abs(max_column(max_disparities))

    # If fairness_threshold is defined, get max between threshold and max absolute disparity
    if fairness_threshold is not None:
        x_domain_limit = math.ceil(max(abs_max_disparity, fairness_threshold))
    else:
        x_domain_limit = math.ceil(abs_max_disparity)

    x_domain = [-x_domain_limit, x_domain_limit]
    position_scales["x"] = alt.Scale(domain=x_domain, range=x_range)

    # METRICS SCALE
    y_range = get_chart_size_range(chart_height, Disparity_Chart.padding_y)
    if chart_height < 300:
        y_range[0] = 30
    y_domain = [metric.upper() for metric in metrics]
    position_scales["y"] = alt.Scale(domain=y_domain, range=y_range)

    return position_scales


def __draw_metrics_rules(metrics, scales, concat_chart):
    """Draws an horizontal rule and the left-hand side label for each metric.
    The groups' bubbles will be positioned on this horizontal rule."""

    metrics_labels = [metric.upper() for metric in metrics]

    metrics_axis = alt.Axis(
        domain=False,
        ticks=False,
        orient="left",
        labelAngle=Metric_Axis.label_angle,
        # LabelPadding logic:
        # Spaces the labels further from the chart if they are part of a concatenated chart
        labelPadding=Metric_Axis.label_padding
        if not concat_chart
        else Metric_Axis.label_padding_concat_chart,
        title="",
    )

    rules_df = pd.DataFrame(
        {
            "metric": metrics_labels,
            "x": scales["x"]["domain"][0],
            "x2": scales["x"]["domain"][1],
        }
    )

    metrics_rules = (
        alt.Chart(rules_df)
        .mark_rule(
            strokeWidth=Metric_Axis.stroke_width,
            stroke=Metric_Axis.stroke,
            tooltip="",
        )
        .encode(
            y=alt.Y("metric:N", scale=scales["y"], axis=metrics_axis),
            x=alt.X("x:Q", scale=scales["x"]),
            x2="x2:Q",
        )
    )

    return metrics_rules


def __get_x_axis_values(x_domain, zero=True):
    TICK_STEP_OPTIONS = [1, 2, 5, 10, 20, 50, 100]

    def list_axis_values(limit, step):
        axis_start = max([1, step - 1])
        positive_axis_values = list(range(axis_start, limit, step))
        negative_axis_values = [-x for x in positive_axis_values][::-1]
        axis_values = positive_axis_values + negative_axis_values
        return axis_values

    domain_limit = x_domain[1] + 1

    for tick_step in TICK_STEP_OPTIONS:
        if domain_limit / tick_step <= 6 or tick_step == TICK_STEP_OPTIONS[-1]:
            axis_values = list_axis_values(domain_limit, tick_step)
            break

    if zero:
        return axis_values + [0]
    return axis_values


def __draw_x_ticks_labels(scales, chart_height):
    """Draws the numbers in the horizontal axis."""

    # The values to be drawn, we don't want to draw 0 (which corresponds to a ratio of 1) as we later draw an annotation.
    axis_values = __get_x_axis_values(scales["x"].domain)

    # Given the semantic of the chart, (how many times smaller or larger) we draw absolute values.
    axis_values_labels = [abs(x) + 1 if x != 0 else "=" for x in axis_values]

    axis_df = pd.DataFrame({"value": axis_values, "label": axis_values_labels})

    tick_labels = (
        alt.Chart(axis_df)
        .mark_text(
            tooltip="",
            align="center",
            font=FONT,
            fontSize=Axis.label_font_size,
            fontWeight=Axis.label_font_weight,
            color=Axis.label_color,
        )
        .encode(
            text=alt.Text("label:N"),
            x=alt.X(
                "value:Q",
                scale=scales["x"],
            ),
            y=alt.value(Disparity_Chart.padding_y * chart_height * 0.7),
        )
    )

    return tick_labels


def __draw_text_annotations(ref_group, chart_height, x_range):
    """Draws on chart text annotations."""

    # FONT
    annotation_text_params = dict(
        font=FONT,
        fontWeight=Annotation.font_weight,
        tooltip="",
    )

    # TIMES LARGER TEXT
    text_times_larger = (
        alt.Chart(DUMMY_DF)
        .mark_text(
            align="right",
            fill=Annotation.font_color,
            fontSize=Annotation.font_size,
            **annotation_text_params,
        )
        .encode(
            x=alt.value(x_range[1]),
            y=alt.value(Disparity_Chart.padding_y * chart_height * 0.3),
            text=alt.value("Times Larger"),
        )
    )

    # TIMES SMALLER TEXT
    text_times_smaller = (
        alt.Chart(DUMMY_DF)
        .mark_text(
            align="left",
            fill=Annotation.font_color,
            fontSize=Annotation.font_size,
            **annotation_text_params,
        )
        .encode(
            x=alt.value(x_range[0]),
            y=alt.value(Disparity_Chart.padding_y * chart_height * 0.3),
            text=alt.value("Times Smaller"),
        )
    )

    # EQUAL TEXT
    text_reference_group = (
        alt.Chart(DUMMY_DF)
        .mark_text(
            align="center",
            fontSize=Annotation.font_size,
            fill=Annotation.font_color,
            **annotation_text_params,
        )
        .encode(
            x=alt.value(x_range[0] + (x_range[1] - x_range[0]) / 2),
            y=alt.value(Disparity_Chart.padding_y * chart_height * 0.3),
            text=alt.value("Equal"),
        )
    )

    return text_times_smaller + text_times_larger + text_reference_group


def __draw_reference_rule(ref_group, chart_height, chart_width):
    """Draws vertical reference rule where the ratio is the same as the reference group."""

    reference_rule = (
        alt.Chart(DUMMY_DF)
        .mark_rule(
            strokeWidth=Reference_Rule.stroke_width,
            stroke=Reference_Rule.stroke,
            strokeDash=Reference_Rule.stroke_dash,
        )
        .encode(
            x=alt.value(chart_width / 2),
            y=alt.value(chart_height * Disparity_Chart.padding_y / 1.2),
            y2=alt.value(chart_height * (1 - Disparity_Chart.padding_y / 1.2)),
            tooltip=alt.value(f"{ref_group} [REF]"),
        )
    )

    return reference_rule


def __draw_threshold_rules(
    threshold_df, scales, chart_height, accessibility_mode=False
):
    """Draws threshold rules: red lines that mark the defined fairness_threshold in the chart."""
    stroke_color = (
        Threshold_Rule.stroke_accessible
        if accessibility_mode
        else Threshold_Rule.stroke
    )

    threshold_rule = (
        alt.Chart(threshold_df)
        .mark_rule(
            stroke=stroke_color,
            opacity=Threshold_Rule.opacity,
            strokeWidth=Threshold_Rule.stroke_width,
            tooltip="",
        )
        .encode(
            y=alt.value(Disparity_Chart.padding_y * chart_height),
            y2=alt.value((1 - Disparity_Chart.padding_y) * chart_height),
        )
    )

    lower_threshold_rule = threshold_rule.encode(
        x=alt.X("min:Q", scale=scales["x"]),
    )
    upper_threshold_rule = threshold_rule.encode(
        x=alt.X("max:Q", scale=scales["x"]),
    )

    return lower_threshold_rule + upper_threshold_rule


def __draw_threshold_bands(
    threshold_df,
    scales,
    chart_height,
    chart_width,
    accessibility_mode=False,
):
    """Draws threshold bands: regions painted red where the metric value is above the defined fairness_threshold."""
    fill_color = (
        Threshold_Band.color_accessible if accessibility_mode else Threshold_Band.color
    )

    threshold_band = (
        alt.Chart(threshold_df)
        .mark_rect(fill=fill_color, opacity=Threshold_Band.opacity, tooltip="")
        .encode(
            y=alt.value(Disparity_Chart.padding_y * chart_height),
            y2=alt.value((1 - Disparity_Chart.padding_y) * chart_height),
        )
    )

    lower_threshold_band = threshold_band.encode(
        x2="lower_end:Q",
        x=alt.X("min:Q", scale=scales["x"]),
    )
    upper_threshold_band = threshold_band.encode(
        x=alt.X("max:Q", scale=scales["x"]),
        x2="upper_end:Q",
    )

    return lower_threshold_band + upper_threshold_band


def __draw_threshold_text(
    fairness_threshold, ref_group, chart_height, accessibility_mode=False
):
    """Draw text that helps to read the threshold elements of the chart."""
    font_color = (
        Annotation.font_color if accessibility_mode else Annotation.font_color_threshold
    )

    threshold_text = (
        alt.Chart(DUMMY_DF)
        .mark_text(
            baseline="top",
            align="left",
            font=FONT,
            fill=font_color,
            fontSize=Annotation.font_size,
            fontWeight=Annotation.font_weight,
            tooltip="",
        )
        .encode(
            x=alt.value(0),
            y=alt.value(chart_height * (1 - 2 / 3 * Disparity_Chart.padding_y)),
            text=alt.value(
                f"The metric value for any group should not be {fairness_threshold} (or more) times smaller or larger than that of the reference group {ref_group}."
            ),
        )
    )

    return threshold_text


def __get_threshold_elements(
    fairness_threshold,
    ref_group,
    scales,
    chart_height,
    chart_width,
    accessibility_mode=False,
):
    """Gets threshold elements (rules, bands and text) for the chart."""

    # CREATE THRESHOLD DF
    threshold_df = pd.DataFrame(
        {
            "min": -fairness_threshold + 1,
            "max": fairness_threshold - 1,
            "lower_end": scales["x"]["domain"][0],
            "upper_end": scales["x"]["domain"][1],
        },
        index=[0],
    )

    # RULES
    threshold_rules = __draw_threshold_rules(
        threshold_df, scales, chart_height, accessibility_mode
    )

    # BANDS
    threshold_bands = __draw_threshold_bands(
        threshold_df,
        scales,
        chart_height,
        chart_width,
        accessibility_mode,
    )

    # HELPER TEXT
    threshold_text = __draw_threshold_text(
        fairness_threshold, ref_group, chart_height, accessibility_mode
    )

    return threshold_rules + threshold_bands + threshold_text


def __draw_bubbles(
    plot_table,
    metrics,
    ref_group,
    scales,
    selection,
):
    """Draws the bubbles for all metrics."""

    # X AXIS GRIDLINES
    axis_values = __get_x_axis_values(scales["x"].domain, zero=False)

    x_axis = alt.Axis(
        values=axis_values, ticks=False, domain=False, labels=False, title=None
    )

    # COLOR
    bubble_color_encoding = alt.condition(
        selection,
        alt.Color("attribute_value:N", scale=scales["color"], legend=None),
        alt.value(Bubble.color_faded),
    )

    plot_table["tooltip_group_size"] = plot_table.apply(
        lambda row: get_tooltip_text_group_size(
            row["group_size"], row["total_entities"]
        ),
        axis=1,
    )

    # CHART INITIALIZATION
    bubble_centers = alt.Chart().mark_point()
    bubble_areas = alt.Chart().mark_circle()

    # LAYERING THE METRICS
    for metric in metrics:

        plot_table[f"tooltip_disparity_explanation_{metric}"] = plot_table.apply(
            lambda row: get_tooltip_text_disparity_explanation(
                row[f"{metric}_disparity_scaled"],
                row["attribute_value"],
                metric,
                ref_group,
            ),
            axis=1,
        )

        bubble_tooltip_encoding = [
            alt.Tooltip(field="attribute_value", type="nominal", title="Group"),
            alt.Tooltip(field="tooltip_group_size", type="nominal", title="Group Size"),
            alt.Tooltip(
                field=f"tooltip_disparity_explanation_{metric}",
                type="nominal",
                title="Disparity",
            ),
            alt.Tooltip(
                field=f"{metric}",
                type="quantitative",
                format=".2f",
                title=f"{metric}".upper(),
            ),
        ]

        # BUBBLE CENTERS
        trigger_centers = alt.selection_multi(empty="all", fields=["attribute_value"])

        bubble_centers += (
            alt.Chart(plot_table)
            .transform_calculate(metric_variable=f"'{metric.upper()}'")
            .mark_point(filled=True, size=Bubble.center_size)
            .encode(
                x=alt.X(f"{metric}_disparity_scaled:Q", scale=scales["x"], axis=x_axis),
                y=alt.Y("metric_variable:N", scale=scales["y"], axis=no_axis()),
                tooltip=bubble_tooltip_encoding,
                color=bubble_color_encoding,
                shape=alt.Shape(
                    "attribute_value:N", scale=scales["shape"], legend=None
                ),
            )
            .add_selection(trigger_centers)
        )

        # BUBBLE AREAS
        trigger_areas = alt.selection_multi(empty="all", fields=["attribute_value"])

        bubble_areas += (
            alt.Chart(plot_table)
            .mark_circle(opacity=Bubble.opacity)
            .transform_calculate(metric_variable=f"'{metric.upper()}'")
            .encode(
                x=alt.X(f"{metric}_disparity_scaled:Q", scale=scales["x"], axis=x_axis),
                y=alt.Y("metric_variable:N", scale=scales["y"], axis=no_axis()),
                tooltip=bubble_tooltip_encoding,
                color=bubble_color_encoding,
                size=alt.Size("group_size:Q", legend=None, scale=scales["bubble_size"]),
            )
            .add_selection(trigger_areas)
        )

    return bubble_areas + bubble_centers


def get_disparity_bubble_chart_components(
    plot_table,
    metrics,
    ref_group,
    global_scales,
    selection,
    fairness_threshold,
    chart_height,
    chart_width,
    accessibility_mode=False,
    concat_chart=False,
):
    """Creates the components necessary to plot the disparity chart."""

    # COMPUTE SCALED DISPARITIES
    for metric in metrics:
        plot_table[f"{metric}_disparity_scaled"] = plot_table.apply(
            lambda row: transform_ratio(row[f"{metric}_disparity"]), axis=1
        )

    # POSITION SCALES
    position_scales = __get_position_scales(
        plot_table, metrics, fairness_threshold, chart_height, chart_width
    )

    scales = dict(global_scales, **position_scales)

    # RULES
    horizontal_rules = __draw_metrics_rules(metrics, scales, concat_chart)
    reference_rule = __draw_reference_rule(ref_group, chart_height, chart_width)

    # LABELS
    x_ticks_labels = __draw_x_ticks_labels(scales, chart_height)

    # ANNOTATIONS
    text_annotations = __draw_text_annotations(
        ref_group, chart_height, scales["x"].range
    )

    # BUBBLES - CENTERS & AREAS
    bubbles = __draw_bubbles(
        plot_table,
        metrics,
        ref_group,
        scales,
        selection,
    )

    # THRESHOLD & BANDS
    if fairness_threshold is not None:
        threshold_elements = __get_threshold_elements(
            fairness_threshold,
            ref_group,
            scales,
            chart_height,
            chart_width,
            accessibility_mode,
        )

        # ASSEMBLE CHART WITH THRESHOLD
        main_chart = (
            horizontal_rules
            + reference_rule
            + x_ticks_labels
            + text_annotations
            + threshold_elements
            + bubbles
        )

    # ASSEMBLE CHART WITHOUT THRESHOLD
    else:
        main_chart = (
            horizontal_rules
            + x_ticks_labels
            + text_annotations
            + reference_rule
            + bubbles
        )

    return main_chart


def plot_disparity_bubble_chart(
    disparity_df,
    metrics_list,
    attribute,
    fairness_threshold=1.25,
    chart_height=None,
    chart_width=Disparity_Chart.full_width,
    accessibility_mode=False,
):
    """Draws bubble chart to visualize disparity in selected metrics versus that of a
    reference group for a given attribute.

    :param disparity_df: a dataframe generated by the Aequitas Bias class
    :type disparity_df: pandas.core.frame.DataFrame
    :param metrics_list: a list of the metrics of interest
    :type metrics_list: list
    :param attribute: an attribute to plot
    :type attribute: str
    :param fairness_threshold: a value for the maximum allowed disparity, defaults to 1.25
    :type fairness_threshold: float, optional
    :param chart_height: a value (in pixels) for the height of the chart
    :type chart_height: int, optional
    :param chart_width: a value (in pixels) for the width of the chart
    :type chart_width: int, optional
    :param accessibility_mode: a switch for the display of more accessible visual elements, defaults to False
    :type accessibility_mode: bool, optional

    :return: the full disparities chart
    :rtype: Altair chart object
    """

    (
        plot_table,
        metrics,
        ref_group,
        global_scales,
        chart_height,
        chart_width,
        selection,
    ) = Initializer.prepare_bubble_chart(
        disparity_df,
        metrics_list,
        attribute,
        fairness_threshold,
        chart_height,
        chart_width,
        Disparity_Chart,
        accessibility_mode,
    )

    # GET MAIN CHART COMPONENTS
    main_chart = get_disparity_bubble_chart_components(
        plot_table,
        metrics,
        ref_group,
        global_scales,
        selection,
        fairness_threshold,
        chart_height,
        chart_width,
        accessibility_mode,
    )

    # ADD LEGEND
    legend = draw_legend(global_scales, selection, chart_width)
    full_chart = main_chart + legend

    # FINALIZE CHART
    disparity_chart = (
        full_chart.configure_view(
            strokeWidth=0,
        )
        .configure_axisLeft(
            labelFontSize=Metric_Axis.label_font_size,
            labelColor=Metric_Axis.label_color,
            labelFont=FONT,
        )
        .configure_title(
            align="center",
            baseline="middle",
            font=FONT,
            fontWeight=Chart_Title.font_weight,
            fontSize=Chart_Title.font_size,
            color=Chart_Title.font_color,
        )
        .properties(
            height=chart_height,
            width=chart_width,
            title=f"Disparities on {attribute.title()}",
            padding=Disparity_Chart.full_chart_padding,
        )
        .resolve_scale(y="independent", size="independent")
    )

    return disparity_chart



from aequitas.plot.commons.helpers import format_number


def get_tooltip_text_group_size(group_size, total_entities):
    return f'{format_number(group_size)} ({"{0:.2f}".format(group_size / total_entities * 100)}%)'


def get_tooltip_text_disparity_explanation(
    disparity_value, attribute_value, metric, ref_group
):
    if attribute_value == ref_group:
        return "Reference group"

    if disparity_value > 0:
        return (
            "{0:.2f}".format(disparity_value + 1)
            + " times larger "
            + metric.upper()
            + " than the reference group"
        )
    elif disparity_value < 0:
        return (
            "{0:.2f}".format(-disparity_value + 1)
            + " times smaller "
            + metric.upper()
            + " than the reference group"
        )
    else:
        return f"Same {metric.upper()} as the reference group"


def get_tooltip_text_parity_test_explanation(parity_result, metric, fairness_threshold):
    if parity_result == "Reference":
        return "Reference group"
    if parity_result == "Pass":
        return f"Pass. Below fairness threshold of {fairness_threshold}"
    if parity_result == "Fail":
        return f"Fail. Above fairness threshold of {fairness_threshold}"
